"""
Common utilities and base classes.

"""
from abc import ABC, abstractmethod
from collections import OrderedDict
from typing import Any, Dict, Iterable, Iterator, List, Type, Union

import pandas as pd


class ItemBase:
    """
    Base class for representing an item.

    """

    # ID of the Item
    _id: str

    # Dictionary of attributes associated with the Item
    _attributes: dict

    def __init__(self, id: str, **attributes):
        """
        Initialize item's representational state.

        """
        self._id = id

        attributes["id"] = id
        self._attributes = attributes

    @property
    def id(self) -> str:
        """
        ID of the item.

        """
        return self._id

    def __eq__(self, right):
        """
        Two items are equal if they are the same type and have the same ID

        """
        # intentionally comparing types instead of using isinstance(), because
        # we don't want to consider a subclass to be equal
        return type(self) == type(right) and self.id == right.id  # noqa: E721

    def __getitem__(self, attr: str):
        """
        Get the value of any attribute in self.attributes.

        """
        return self._attributes.get(attr)

    def get(self, attr: str, default: Any = None):
        """
        Get the value of any attribute in self.attributes.

        """
        return self._attributes.get(attr, default)

    def __repr__(self):
        """
        Format the item representation.

        """
        if hasattr(self, "name"):
            return f"{self.__class__.__name__}" f'(id="{self.id}", name="{self.name}")'
        else:
            return f'{self.__class__.__name__}(id="{self.id}")'

    def to_dict(self) -> dict:
        """
        Dictionary representation of the item's attributes.

        """
        return self._attributes


class ItemSet(ABC):
    """
    Base class for representing a set of items.

    """

    # The Items in this set. Maps id to item.
    _items: Dict[str, ItemBase]

    def __init__(self, items: Iterable[ItemBase] = ()):
        """
        Initialize items set representational state.

        """
        self._items = OrderedDict()
        for item in items:
            self.add(item)

    def __iter__(self) -> Iterator[ItemBase]:
        """
        Iterate over the items of the set.

        """
        return iter(self._items.values())

    def __len__(self):
        """
        Number of items in the set.

        """
        return len(self._items)

    def __getitem__(self, id: str):
        """
        Get an item.

        """
        return self._items.get(id)

    @property
    @abstractmethod
    def _item_class(self) -> Type[ItemBase]:
        """
        Instance type of items in this set.

        """
        pass

    def get(self, id: str) -> ItemBase:
        """
        Get an item by id.

        Raises:
            ValueError: if the set does not contain an item with the ID.

        """
        try:
            return self._items[id]
        except KeyError:
            pass

        raise ValueError(
            f"{self.__class__.__name__} does not contain "
            f"{self._item_class.__name__} with ID {id}"
        )

    def add(self, item: ItemBase):
        """
        Add a single item to the set. Must be the same type as other members
        of the collection

        """
        if type(item) is not self._item_class:
            raise TypeError(
                f"cannot add {str(item)}; must be type " f"{self._item_class.__name__}"
            )

        self._items[item.id] = item

    def update(self, items: Iterable[ItemBase]):
        """
        Add an iterable of item(s) to this set. All items must be the same
        class as members of this collection.

        """
        for item in items:
            if type(item) is not self._item_class:
                raise TypeError(
                    f"cannot update with {str(item)}; all items must be type "
                    f"{self._item_class.__name__}"
                )

            self._items[item.id] = item

    def remove(self, *items: Union[str, ItemBase]):
        """
        Remove item(s) from this set.

        """
        for item in items:
            if type(item) is str:
                # item is an id
                self._items.pop(item, None)
            else:
                # item is an instance of ItemBase
                self._items.pop(item.id, None)

    def ids(self) -> Iterator[str]:
        """
        Iterator over the IDs of the items in this set.

        """
        return iter(self._items.keys())

    def __repr__(self):
        """
        String representation of the set of items. Returns information on
        the number of items in the set, the type of the set, and up to 3
        items.

        """
        item_strs = []
        for i, item in enumerate(self):
            if i > 2:
                item_strs.append(f"\t... (and {len(self._items.keys()) - 3} others)\n")
                break

            item_strs.append(f"\t{str(item)}\n")

        items_str = "".join(item_strs)
        return f"{type(self).__name__} {{\n{items_str}}}"

    def to_list(self) -> List[dict]:
        """
        List of all the items in the set. Each item is formatted as a
        dictionary, using the item's `to_dict()` method.

        """
        items_list = []
        for item in self._items.values():
            items_list.append(item.to_dict())

        return items_list

    def to_dataframe(self) -> pd.DataFrame:
        """Convert items to a dataframe (wraps `to_list()`)"""
        return pd.DataFrame(self.to_list())



from typing import Dict, List
from unittest import TestCase, mock

import numpy as np

from runeq import Config, errors, stream


def mock_get_json_response(
    bodies: List[dict],
    calls: List,
    status_code=200,
    headers: List[Dict[str, str]] = None,
):
    """Return a function that can be used to mock .get_json_response()

    Args:
        bodies: list of JSON bodies to return from response.json()
        calls: list. Every time the function is called, args and kwargs will
            be appended to this list.
        status_code: status code to return for each response
        headers: the response header to apply
    """
    headers = headers or [{}] * len(bodies)

    num = 0

    def _func(*args, **kwargs):
        nonlocal num
        # add inputs to the list of calls that was provided
        calls.append((args, kwargs))

        resp = mock.MagicMock()
        resp.headers = headers[num]
        resp.status_code = status_code
        resp.ok = status_code < 400
        resp.json.return_value = bodies[num]
        num += 1
        return resp

    return _func


def mock_get_csv_response(
    bodies: List[str],
    calls: List,
    status_code=200,
    headers: List[Dict[str, str]] = None,
):
    """Return a function that can be used to mock .get_csv_response()

    Args:
        bodies: list of response bodies to return from response.text
        calls: list. Every time the function is called, args and kwargs will
            be appended to this list.
        status_code: status code to return for each response
        headers: the response header to apply
    """
    headers = headers or [{}] * len(bodies)

    num = 0

    def _func(*args, **kwargs):
        nonlocal num
        # keep track of the kwargs that were used to call this
        calls.append((args, kwargs))

        resp = mock.MagicMock()
        resp.headers = headers[num]
        resp.status_code = status_code
        resp.ok = status_code < 400
        resp.text = bodies[num]
        num += 1
        return resp

    return _func


class TestStreamV1Client(TestCase):
    """
    Test stream.V1Client and the associated accessors.

    """

    def setUp(self) -> None:
        """
        Initialize a client, set up basic mocking.

        """
        self.cfg = Config(
            client_key_id="abc",
            client_access_key="abc123",
        )
        self.client = stream.V1Client(self.cfg)
        self.use_np_orig = stream.v1.USE_NUMPY
        stream.v1.USE_NUMPY = False

    def tearDown(self) -> None:
        """
        Tear down monkey-patching.

        """
        stream.v1.USE_NUMPY = self.use_np_orig

    @mock.patch("runeq.stream.v1.requests")
    def test_get_json_response(self, requests):
        """
        Test the signature of JSON requests.

        """
        for test_num, case in enumerate(
            (
                (self.client.Accel, "/v1/accel.json"),
                (self.client.BandPower, "/v1/band_power.json"),
                (self.client.Event, "/v1/event.json"),
                (self.client.HeartRate, "/v1/heartrate.json"),
                (self.client.LFP, "/v1/lfp.json"),
                (self.client.ProbabilitySymptom, "/v1/probability_symptom.json"),
                (self.client.Rotation, "/v1/rotation.json"),
                (self.client.Span, "/v1/span.json"),
                (self.client.State, "/v1/state.json"),
            )
        ):
            resource_creator, endpoint = case
            resource = resource_creator(leslie="knope")
            resource.get_json_response(ron="swanson", test_num=test_num)
            requests.get.assert_has_calls(
                [
                    mock.call(
                        self.cfg.stream_url + endpoint,
                        headers=self.cfg.auth_headers,
                        params={
                            "leslie": "knope",
                            "ron": "swanson",
                            "test_num": test_num,
                        },
                    ),
                ]
            )

    @mock.patch("runeq.stream.v1.requests")
    def test_get_csv_response(self, requests):
        """
        Test the signature of CSV requests.

        """
        for test_num, case in enumerate(
            (
                (self.client.Accel, "/v1/accel.csv"),
                (self.client.BandPower, "/v1/band_power.csv"),
                (self.client.HeartRate, "/v1/heartrate.csv"),
                (self.client.LFP, "/v1/lfp.csv"),
                (self.client.ProbabilitySymptom, "/v1/probability_symptom.csv"),
                (self.client.Rotation, "/v1/rotation.csv"),
                (self.client.State, "/v1/state.csv"),
            )
        ):
            resource_creator, endpoint = case
            resource = resource_creator(leslie="knope")
            resource.get_csv_response(ron="swanson", test_num=test_num)
            requests.get.assert_has_calls(
                [
                    mock.call(
                        self.cfg.stream_url + endpoint,
                        stream=True,
                        headers=self.cfg.auth_headers,
                        params={
                            "leslie": "knope",
                            "ron": "swanson",
                            "test_num": test_num,
                        },
                    ),
                ]
            )

    def test_iter_json_data_with_token(self):
        """
        Test the iterator over JSON responses, paginating with the next page
        token header.

        """
        for test_num, resource_creator in enumerate(
            (
                self.client.Accel,
                self.client.BandPower,
                self.client.Event,
                self.client.HeartRate,
                self.client.LFP,
                self.client.ProbabilitySymptom,
                self.client.Rotation,
                self.client.Span,
                self.client.State,
            )
        ):
            resource = resource_creator()

            mock_responses = [
                {"success": True, "result": [], "next_page": 1},
                {"success": True, "result": []},
            ]

            calls = []
            resource.get_json_response = mock_get_json_response(
                mock_responses,
                calls,
                200,
                [
                    {"X-Rune-Next-Page-Token": "MTIzNDU2MDAwMA=="},
                    {},
                ],
            )

            iterator = resource.iter_json_data(test_num=test_num)
            self.assertEqual(len(list(iterator)), 2)

            # Check that all parameters were kept the same across calls,
            # except for "next_page_token"
            self.assertEqual(
                calls,
                [
                    ((), {"test_num": test_num}),
                    ((), {"test_num": test_num, "next_page_token": "MTIzNDU2MDAwMA=="}),
                ],
            )

    def test_iter_json_data(self):
        """
        Test the iterator over JSON responses, following pagination with
        the page number.

        """
        results = [{"a": 1}, {"b": 2}]
        mock_responses = [
            {"success": True, "result": results[0], "next_page": 1},
            {"success": True, "result": results[1]},
        ]

        for test_num, resource_creator in enumerate(
            (
                self.client.Accel,
                self.client.BandPower,
                self.client.Event,
                self.client.HeartRate,
                self.client.LFP,
                self.client.ProbabilitySymptom,
                self.client.Rotation,
                self.client.State,
            )
        ):
            resource = resource_creator()

            #
            # Successful Requests
            #
            calls = []
            resource.get_json_response = mock_get_json_response(mock_responses, calls)

            # Check the results
            num_results = 0
            iterator = resource.iter_json_data(test_num=test_num)
            for i, actual in enumerate(iterator):
                self.assertEqual(results[i], actual)
                num_results += 1

            self.assertEqual(num_results, 2)

            # Check that all parameters were kept the same across calls,
            # except for "page" (which must be incremented)
            self.assertEqual(
                calls,
                [((), {"test_num": test_num}), ((), {"test_num": test_num, "page": 1})],
            )

            #
            # Request Error
            # Iterator should check the response status for each request
            #
            err_details = {
                "message": "i am an intentional error!",
                "type": "TestError",
            }
            resource.get_json_response = mock_get_json_response(
                [{"success": False, "error": err_details}],
                [],
                status_code=404,
            )

            with self.assertRaises(errors.APIError) as e:
                next(resource.iter_json_data())

            err = e.exception
            self.assertEqual(err.status_code, 404)
            self.assertEqual(err.details, err_details)

    def test_iter_csv_data_with_token(self):
        """
        Test the iterator over CSV responses, which follows new pagination

        """
        mock_responses = [
            "good,better\nskiing,hiking\n",
            "good,better\ncupcakes,brownies\n",
            "",
        ]

        for test_num, resource_creator in enumerate(
            (
                self.client.Accel,
                self.client.BandPower,
                self.client.HeartRate,
                self.client.LFP,
                self.client.ProbabilitySymptom,
                self.client.Rotation,
                self.client.State,
            )
        ):
            resource = resource_creator()

            #
            # Successful Requests
            #
            calls = []
            resource.get_csv_response = mock_get_csv_response(
                mock_responses,
                calls,
                200,
                [
                    {"X-Rune-Next-Page-Token": "MTIzNDU2MDAwMA=="},
                    {"X-Rune-Next-Page-Token": "MTIzNDU2MDAwMA=="},
                    {},
                ],
            )

            # Check the results
            iterator = resource.iter_csv_text(test_num=test_num)
            self.assertEqual(len(list(iterator)), 2)

            # Check that all parameters were kept the same across calls,
            # except for "next_page_token" (which will normally be different
            # for each response)
            self.assertEqual(
                calls,
                [
                    ((), {"test_num": test_num}),
                    ((), {"test_num": test_num, "next_page_token": "MTIzNDU2MDAwMA=="}),
                    ((), {"test_num": test_num, "next_page_token": "MTIzNDU2MDAwMA=="}),
                ],
            )

    def test_iter_csv_data(self):
        """
        Test the iterator over CSV responses, which follows pagination
        """
        mock_responses = [
            "good,better\nskiing,hiking\n",
            "good,better\ncupcakes,brownies\n",
            "",
        ]

        for test_num, resource_creator in enumerate(
            (
                self.client.Accel,
                self.client.BandPower,
                self.client.HeartRate,
                self.client.LFP,
                self.client.ProbabilitySymptom,
                self.client.Rotation,
                self.client.State,
            )
        ):
            resource = resource_creator()

            #
            # Successful Requests
            #
            calls = []
            resource.get_csv_response = mock_get_csv_response(
                mock_responses,
                calls,
            )

            # Check the results
            num_results = 0
            iterator = resource.iter_csv_text(test_num=test_num)
            for i, actual in enumerate(iterator):
                self.assertEqual(mock_responses[i], actual)
                num_results += 1

            # although there are 3 responses, the last (empty) body should not
            # be returned by the iterator
            self.assertEqual(num_results, 2)

            # Check that all parameters were kept the same across calls,
            # except for "page" (which must be incremented)
            self.assertEqual(
                calls,
                [
                    ((), {"test_num": test_num}),
                    ((), {"test_num": test_num, "page": 1}),
                    ((), {"test_num": test_num, "page": 2}),
                ],
            )

            #
            # Request Error
            # Iterator should check the response status for each request
            #
            err_details = {
                "message": "i am an intentional error!",
                "type": "TestError",
            }
            # note: CSV endpoints return JSON on API errors
            resource.get_csv_response = mock_get_json_response(
                [{"success": False, "error": err_details}],
                [],
                status_code=404,
            )

            with self.assertRaises(errors.APIError) as e:
                next(resource.iter_csv_text())

            err = e.exception
            self.assertEqual(err.status_code, 404)
            self.assertEqual(err.details, err_details)

    def test_iter_points(self):
        """
        Test iterating over data as points. Uses the CSV endpoint.

        """
        mock_responses = [
            "lower,higher,label\n1,2,ints\n3.5,6.7,floats\n",
            "lower,higher,label\n,8.9,missing data\n",
            "",
        ]
        expected = [
            {"lower": 1, "higher": 2, "label": "ints"},
            {"lower": 3.5, "higher": 6.7, "label": "floats"},
            {"lower": None, "higher": 8.9, "label": "missing data"},
        ]

        for test_num, resource_creator in enumerate(
            (
                self.client.Accel,
                self.client.BandPower,
                self.client.HeartRate,
                self.client.LFP,
                self.client.ProbabilitySymptom,
                self.client.Rotation,
                self.client.State,
            )
        ):
            resource = resource_creator()
            # replace get_csv_response on the resource
            resource.get_csv_response = mock_get_csv_response(
                mock_responses,
                [],
            )

            for i, point in enumerate(resource.points()):
                self.assertDictEqual(expected[i], point)

            # replace get_csv_response again, to restart the mock responses
            resource.get_csv_response = mock_get_csv_response(
                mock_responses,
                [],
            )

            for i, point in enumerate(resource):
                self.assertDictEqual(expected[i], point)
                # check dtype for "higher", which always has a numeric
                # value in the test data
                self.assertNotIsInstance(point["higher"], np.float64)

    def test_iter_points_numpy(self):
        """
        Test iterating over data as points, using Numpy to convert.

        """
        stream.v1.USE_NUMPY = True

        mock_responses = [
            "lower,higher,label\n1,2,ints\n3.5,6.7,floats\n",
            "lower,higher,label\n,8.9,missing data\n",
            "",
        ]
        expected = [
            {"lower": 1, "higher": 2, "label": "ints"},
            {"lower": 3.5, "higher": 6.7, "label": "floats"},
            {"lower": np.NaN, "higher": 8.9, "label": "missing data"},
        ]

        for test_num, resource_creator in enumerate(
            (
                self.client.Accel,
                self.client.BandPower,
                self.client.HeartRate,
                self.client.LFP,
                self.client.ProbabilitySymptom,
                self.client.Rotation,
                self.client.State,
            )
        ):
            resource = resource_creator()
            # replace get_csv_response on the resource
            resource.get_csv_response = mock_get_csv_response(
                mock_responses,
                [],
            )

            for i, point in enumerate(resource.points()):
                self.assertDictEqual(expected[i], point)
                # check dtype for "higher", which always has a numeric
                # value in the test data
                self.assertIsInstance(point["higher"], np.float64)



"""
Query data directly from the V2 Stream API.

"""
import time
from datetime import date, datetime
from typing import Iterable, Iterator, Optional, Union

from .client import StreamClient, global_stream_client

_time_type = Union[int, float, date, datetime]


def _time_type_to_unix_secs(t: _time_type) -> Union[int, float]:
    """Standardize time input as timestamp (in unix secs)"""
    if isinstance(t, datetime):
        return t.timestamp()
    elif isinstance(t, date):
        return time.mktime(t.timetuple())
    else:
        return t


def get_stream_data(
    stream_id: str,
    start_time: Optional[_time_type] = None,
    start_time_ns: Optional[int] = None,
    end_time: Optional[_time_type] = None,
    end_time_ns: Optional[int] = None,
    format: Optional[str] = "csv",
    limit: Optional[int] = None,
    page_token: Optional[str] = None,
    timestamp: Optional[str] = "iso",
    timezone: Optional[int] = None,
    timezone_name: Optional[str] = None,
    translate_enums: Optional[bool] = True,
    client: Optional[StreamClient] = None,
) -> Iterator[Union[str, dict]]:
    """
    Fetch raw data for a stream.

    Args:
        stream_id: ID of the stream
        start_time: Start time for the query, provided as a unix timestamp
            (in seconds), a datetime.datetime, or a datetime.date.
        start_time_ns: Start time for the query, provided as a unix timestamp
            (in nanoseconds).
        end_time: End time for the query, provided as a unix timestamp
            (in seconds), a datetime.datetime, or a datetime.date.
        end_time_ns: End time for the query, provided as a unix timestamp
            (in nanoseconds).
        format: Either "csv" (default) or "json". Determines the content type
            of the API response.
        limit: Maximum number of timestamps to return, across *all pages*
            of the response. A limit of 0 (default) will fetch all
            available data.
        page_token: Token to fetch the subsequent page of results.
            The value is obtained from the "X-Rune-Next-Page-Token"
            response header field.
        timestamp: One of "unix", "unixns", or "iso", which determines
            how timestamps are formatted in the response
        timezone: Timezone offset, in seconds, used to calculate
            string-based timestamp formats such as datetime and iso.
            For example, PST (UTC-0800) is represented as -28800.
            If omitted, the timezone is UTC.
        timezone_name: The name from the IANA timezone database used to
            calculate string-based timestamp formats such as datetime and iso.
            Returns the correct UTC offset for a given date/time in order to
            account for daylight savings time.
        translate_enums: If True, enum values are returned as their string
            representation. Otherwise, enums are returned as integer values.
        client: If specified, this client is used to fetch data from the
            API. Otherwise, the global
            :class:`~runeq.resources.client.StreamClient` is used.

    Returns:
        An iterator over paginated API responses. If format is "json", each
        response is a dict. If format is "csv", each response is a
        CSV-formatted string.

    """
    if start_time and start_time_ns:
        raise ValueError("only start_time or start_time_ns can be defined, not both.")

    if end_time and end_time_ns:
        raise ValueError("only end_time or end_time_ns can be defined, not both.")

    start_time = _time_type_to_unix_secs(start_time)
    end_time = _time_type_to_unix_secs(end_time)

    client = client or global_stream_client()
    path = f"/v2/streams/{stream_id}"

    yield from client.get_data(
        path,
        start_time=start_time,
        start_time_ns=start_time_ns,
        end_time=end_time,
        end_time_ns=end_time_ns,
        format=format,
        limit=limit,
        page_token=page_token,
        timestamp=timestamp,
        timezone=timezone,
        timezone_name=timezone_name,
        translate_enums=translate_enums,
    )


def get_stream_availability(
    stream_ids: Union[str, Iterable[str]],
    start_time: _time_type,
    end_time: _time_type,
    resolution: int,
    batch_operation: Optional[str] = None,
    format: Optional[str] = "csv",
    limit: Optional[int] = None,
    page_token: Optional[str] = None,
    timestamp: Optional[str] = "iso",
    timezone: Optional[int] = None,
    timezone_name: Optional[str] = None,
    client: Optional[StreamClient] = None,
) -> Iterator[Union[str, dict]]:
    """
    Fetch the availability of 1 or multiple streams. When multiple stream
    IDs are specified, this fetches the availability of **all** or **any**
    of the streams (depending on the **batch_operation**).

    Args:
        stream_ids: 1 or multiple stream IDs. If multiple stream IDs are
            specified, **batch_operation** is also required.
        start_time: Start time for the query, provided as a unix timestamp
            (in seconds), a datetime.datetime, or a datetime.date.
        end_time: End time for the query, provided as a unix timestamp
            (in seconds), a datetime.datetime, or a datetime.date.
        resolution: Interval between returned timestamps, in seconds.
        batch_operation: Either "any" or "all", which determines
            what type of batch calculation will determine availability for
            multiple streams. Availability values will equal 1 when
            data is available for "all" or "any" of the requested streams in
            the given interval. This argument is required when multiple
            **stream_ids** are specified.
        format: Either "csv" (default) or "json". Determines the content type
            of the API response.
        limit: Maximum number of timestamps to return, across *all pages*
            of the response. A limit of 0 (default) will fetch all
            available data.
        page_token: Token to fetch the subsequent page of results.
            The value is obtained from the "X-Rune-Next-Page-Token"
            response header field.
        timestamp: One of "unix", "unixns", or "iso", which determines
            how timestamps are formatted in the response
        timezone: Timezone offset, in seconds, used to calculate
            string-based timestamp formats such as datetime and iso.
            For example, PST (UTC-0800) is represented as -28800.
            If omitted, the timezone is UTC.
        timezone_name: The name from the IANA timezone database used to
            calculate string-based timestamp formats such as datetime and iso.
            Returns the correct UTC offset for a given date/time in order to
            account for daylight savings time.
        client: If specified, this client is used to fetch data from the
            API. Otherwise, the global
            :class:`~runeq.resources.client.StreamClient` is used.

    Returns:
        An iterator over paginated API responses. If format is "json", each
        response is a dict. If format is "csv", each response is a
        CSV-formatted string.

    Raises:
        ValueError: if batch_operation is not specified and querying for
            more than 1 stream_id

    """
    params = {
        "start_time": _time_type_to_unix_secs(start_time),
        "end_time": _time_type_to_unix_secs(end_time),
        "resolution": resolution,
        "batch_operation": batch_operation,
        "format": format,
        "limit": limit,
        "page_token": page_token,
        "timestamp": timestamp,
        "timezone": timezone,
        "timezone_name": timezone_name,
    }

    client = client or global_stream_client()

    # If stream_ids is not a string, convert it to a list so that we can
    if type(stream_ids) is not str:
        stream_ids = list(stream_ids)

    if type(stream_ids) is str:
        path = f"/v2/streams/{stream_ids}/availability"
    elif len(stream_ids) == 1:
        path = f"/v2/streams/{stream_ids[0]}/availability"
    else:
        # If querying for batch availability, need batch_operation
        if not batch_operation:
            raise ValueError(
                "batch_operation must be specified for multiple stream IDs"
            )

        path = "/v2/batch/availability"
        params["stream_id"] = stream_ids

    yield from client.get_data(path, **params)



import os
from unittest import TestCase, mock

from runeq import Config

TEST_ROOT = os.path.dirname(os.path.realpath(__file__))


class TestConfig(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.access_token_good_config = f"{TEST_ROOT}/data/access_token_good_config.yaml"
        cls.access_token_bad_config = f"{TEST_ROOT}/data/access_token_bad_config.yaml"
        cls.client_keys_good_config = f"{TEST_ROOT}/data/client_keys_good_config.yaml"
        cls.client_keys_bad_config = f"{TEST_ROOT}/data/client_keys_bad_config.yaml"

    def test_init_file(self):
        """
        Test initializing using a yaml file
        """
        # Test access tokens init
        cfg = Config(self.access_token_good_config)
        expected = {
            "X-Rune-User-Access-Token-Id": "foo",
            "X-Rune-User-Access-Token-Secret": "bar",
        }
        self.assertEqual(expected, cfg.auth_headers)
        self.assertEqual(expected, cfg.access_token_auth_headers)

        with self.assertRaises(ValueError):
            _ = cfg.jwt_auth_headers

        with self.assertRaises(ValueError):
            _ = cfg.client_auth_headers

        # Test client keys init
        cfg = Config(self.client_keys_good_config)

        expected = {
            "X-Rune-Client-Key-ID": "abc",
            "X-Rune-Client-Access-Key": "def",
        }
        self.assertEqual(expected, cfg.auth_headers)
        self.assertEqual(expected, cfg.client_auth_headers)
        with self.assertRaises(ValueError):
            _ = cfg.jwt_auth_headers

        with self.assertRaises(ValueError):
            _ = cfg.access_token_auth_headers

        self.assertEqual("https://foo.runelabs.io", cfg.stream_url)

    def test_init_file_invalid(self):
        """
        Test initializing with an invalid YAML file
        """
        # Test invalid access token file
        with self.assertRaises(ValueError):
            Config(self.access_token_bad_config)

        # Test invalid client keys file
        with self.assertRaises(ValueError):
            Config(self.client_keys_bad_config)

    def test_init_kwargs(self):
        """
        Test initializing with valid kwargs
        """
        # access token
        cfg = Config(access_token_id="foo", access_token_secret="bar")
        access_token_headers = {
            "X-Rune-User-Access-Token-Id": "foo",
            "X-Rune-User-Access-Token-Secret": "bar",
        }
        self.assertEqual(access_token_headers, cfg.auth_headers)
        self.assertEqual(access_token_headers, cfg.access_token_auth_headers)
        with self.assertRaises(ValueError):
            _ = cfg.client_auth_headers

        with self.assertRaises(ValueError):
            _ = cfg.jwt_auth_headers

        # client keys
        cfg = Config(
            client_key_id="abc",
            client_access_key="123",
        )
        client_headers = {
            "X-Rune-Client-Key-ID": "abc",
            "X-Rune-Client-Access-Key": "123",
        }
        self.assertEqual(client_headers, cfg.auth_headers)
        self.assertEqual(client_headers, cfg.client_auth_headers)
        with self.assertRaises(ValueError):
            _ = cfg.jwt_auth_headers

        with self.assertRaises(ValueError):
            _ = cfg.access_token_auth_headers

        # jwt
        cfg = Config(jwt="abc123")
        jwt_headers = {"X-Rune-User-Access-Token": "abc123"}
        self.assertEqual(jwt_headers, cfg.auth_headers)
        self.assertEqual(jwt_headers, cfg.jwt_auth_headers)
        with self.assertRaises(ValueError):
            _ = cfg.client_auth_headers

        with self.assertRaises(ValueError):
            _ = cfg.access_token_auth_headers

        # cognito refresh token
        with mock.patch("runeq.config.boto3.client") as boto3_client:
            mock_cognito = mock.Mock()
            mock_cognito.initiate_auth.return_value = {
                "AuthenticationResult": {"AccessToken": "def456"}
            }
            boto3_client.return_value = mock_cognito

            cfg = Config(cognito_refresh_token="ref", cognito_client_id="cli")

        cognito_jwt_headers = {"X-Rune-User-Access-Token": "def456"}
        self.assertEqual(cognito_jwt_headers, cfg.auth_headers)
        self.assertEqual(cognito_jwt_headers, cfg.jwt_auth_headers)
        with self.assertRaises(ValueError):
            _ = cfg.client_auth_headers

        with self.assertRaises(ValueError):
            _ = cfg.access_token_auth_headers

    @mock.patch("runeq.config.boto3.client")
    def test_refresh_auth(self, mock_boto3):
        """Test that refresh_auth updates JWT value and returns a bool"""
        mock_cognito = mock.Mock()
        mock_cognito.initiate_auth.return_value = {
            "AuthenticationResult": {"AccessToken": "jwt1"}
        }
        mock_boto3.return_value = mock_cognito

        # refresh_auth returns False if the Config isn't using a refresh token
        for cfg in (
            Config(client_key_id="abc", client_access_key="123"),
            Config(access_token_id="foo", access_token_secret="bar"),
            Config(jwt="abc123"),
        ):
            self.assertFalse(cfg.refresh_auth())
            mock_cognito.initiate_auth.assert_not_called()

        # with a refresh token, initiate_auth is called once on init
        cfg = Config(cognito_refresh_token="ref", cognito_client_id="cli")
        self.assertEqual(mock_cognito.initiate_auth.call_count, 1)
        self.assertEqual(cfg.auth_headers, {"X-Rune-User-Access-Token": "jwt1"})

        # check that auth headers change after refreshing
        mock_cognito.initiate_auth.return_value = {
            "AuthenticationResult": {"AccessToken": "jwt2"}
        }
        self.assertTrue(cfg.refresh_auth())

        self.assertEqual(mock_cognito.initiate_auth.call_count, 2)
        self.assertEqual(cfg.auth_headers, {"X-Rune-User-Access-Token": "jwt2"})

    def test_init_kwargs_invalid(self):
        """
        Test initializing with invalid kwargs.

        """
        with self.assertRaises(ValueError):
            Config(client_key_id="abc")

        with self.assertRaises(ValueError):
            Config(client_access_key="123")

        with self.assertRaises(ValueError):
            Config(client_access_key="123", jwt="abc123")

        with self.assertRaises(ValueError):
            Config(access_token_id="foo")

        with self.assertRaises(ValueError):
            Config(access_token_id="foo", client_access_key="123")

        with self.assertRaises(ValueError):
            Config(access_token_id="foo", client_access_key="123", jwt="zee")

        with self.assertRaises(ValueError):
            Config(
                jwt="zee",
                cognito_client_id="foo",
                cognito_refresh_token="bar",
            )

    def test_init_invalid(self):
        """
        Cannot initialize with a filename and kwargs
        """
        with self.assertRaises(TypeError):
            Config(self.client_keys_good_config, stream_url="https://bar.runelabs.io")



"""
The Rune Labs Standard Development Kit (SDK) for querying data from Rune
APIs.

"""
from . import resources, stream
from .config import Config
from .resources.client import initialize

__all__ = [
    "Config",
    "initialize",
    "resources",
    "stream",
]



"""
Tests for fetching project data.

"""
from unittest import TestCase, mock

from runeq.config import Config
from runeq.resources.client import GraphClient
from runeq.resources.project import (
    Cohort,
    CohortPatientMetadata,
    CohortSet,
    Metric,
    MetricSet,
    Project,
    ProjectPatientMetadata,
    get_cohort_patients,
    get_project,
    get_project_patients,
    get_projects,
)


class TestProject(TestCase):
    """
    Unit tests for the Project class.

    """

    def setUp(self):
        """
        Set up mock graph client for testing.

        """
        self.mock_client = GraphClient(
            Config(client_key_id="test", client_access_key="config")
        )

    def test_project_attributes(self):
        """
        Test attributes for an initialized Project.

        """
        example_cohorts = CohortSet(
            [
                Cohort(
                    id="cohort-1",
                    title="title-1",
                    created_at=1630515986.9949625,
                    updated_at=1630515986.9949625,
                    created_by="user-1",
                    updated_by="user-1",
                ),
                Cohort(
                    id="cohort-2",
                    title="title-2",
                    created_at=1630515986.9949625,
                    updated_at=1630515986.9949625,
                    created_by="user-1",
                    updated_by="user-1",
                ),
            ]
        )

        test_project = Project(
            id="proj1-id",
            created_at=1630515986.9949625,
            updated_at=1630515986.9949625,
            title="Project 1",
            description="Test description.",
            created_by="user-1",
            updated_by="user-1",
            started_at=1630515986.9949625,
            status="ACTIVE",
            type="SANDBOX",
            cohorts=example_cohorts,
        )

        self.assertEqual("proj1-id", test_project.id)
        self.assertEqual(1630515986.9949625, test_project.created_at)
        self.assertEqual("Project 1", test_project.title)
        self.assertEqual(1630515986.9949625, test_project.updated_at)
        self.assertEqual("user-1", test_project.created_by)
        self.assertEqual("user-1", test_project.updated_by)
        self.assertEqual("ACTIVE", test_project.status)
        self.assertEqual("SANDBOX", test_project.type)
        self.assertEqual(example_cohorts, test_project.cohorts)

    def test_get_project(self):
        """
        Test get project for specified project_id.

        """
        example_cohorts = [
            {
                "id": "test-id-1",
                "title": "test-1",
                "description": "Some cohort 1",
                "created_at": 1667225389.222881,
                "updated_at": 1667225389.084547,
                "created_by": "Computer wizard",
                "updated_by": "Computer wizard",
            }
        ]
        self.mock_client.execute = mock.Mock()
        example_project = {
            "id": "proj1-id",
            "created_at": 1630515986.9949625,
            "updated_at": 1630515986.9949625,
            "title": "Project 1",
            "description": "Test description.",
            "created_by": "user-id-1",
            "updated_by": "user-id-1",
            "started_at": 1630515986.9949625,
            "status": "ACTIVE",
            "type": "SANDBOX",
            "cohortList": {"cohorts": example_cohorts},
        }

        expected_project = {
            "id": "proj1-id",
            "created_at": 1630515986.9949625,
            "updated_at": 1630515986.9949625,
            "title": "Project 1",
            "description": "Test description.",
            "created_by": "user-id-1",
            "updated_by": "user-id-1",
            "started_at": 1630515986.9949625,
            "status": "ACTIVE",
            "type": "SANDBOX",
            "cohorts": example_cohorts,
        }

        self.mock_client.execute.side_effect = [{"project": example_project}]

        project = get_project("proj1-id", client=self.mock_client)

        self.assertEqual(
            expected_project,
            project.to_dict(),
        )

    def test_get_projects_basic(self):
        """
        Test get projects for the initialized user.

        """
        self.mock_client.execute = mock.Mock()
        projects_expected = [
            {
                "id": "proj1-id",
                "created_at": 1630515986.9949625,
                "title": "Project 1",
                "description": "Test description.",
                "created_by": "user-id-1",
                "started_at": 1630515986.9949625,
                "status": "ACTIVE",
                "updated_at": 1630515986.9949625,
                "type": "SANDBOX",
                "updated_by": "user-id-1",
                "cohorts": [],
            },
            {
                "id": "proj2-id",
                "created_at": 1630517987.9949625,
                "title": "Project 2",
                "description": "Test description 2.",
                "created_by": "user-id-2",
                "started_at": 1630517986.9949625,
                "updated_at": 1630515986.9949625,
                "status": "ACTIVE",
                "type": "SANDBOX",
                "updated_by": "user-id-1",
                "cohorts": [],
            },
            {
                "id": "proj3-id",
                "created_at": 1630519988.9949625,
                "title": "Project 3",
                "description": "Test description 3.",
                "created_by": "user-id-3",
                "started_at": 1630519986.9949625,
                "updated_at": 1630515986.9949625,
                "status": "ACTIVE",
                "type": "SANDBOX",
                "updated_by": "user-id-1",
                "cohorts": [],
            },
        ]
        self.mock_client.execute.return_value = {
            "org": {
                "id": "org-rune,org",
                "projectList": {
                    "projects": [
                        {
                            "id": "proj1-id",
                            "created_at": 1630515986.9949625,
                            "title": "Project 1",
                            "description": "Test description.",
                            "created_by": "user-id-1",
                            "started_at": 1630515986.9949625,
                            "status": "ACTIVE",
                            "updated_at": 1630515986.9949625,
                            "type": "SANDBOX",
                            "updated_by": "user-id-1",
                            "cohortList": {"cohorts": []},
                        },
                        {
                            "id": "proj2-id",
                            "created_at": 1630517987.9949625,
                            "title": "Project 2",
                            "description": "Test description 2.",
                            "created_by": "user-id-2",
                            "started_at": 1630517986.9949625,
                            "updated_at": 1630515986.9949625,
                            "status": "ACTIVE",
                            "type": "SANDBOX",
                            "updated_by": "user-id-1",
                            "cohortList": {"cohorts": []},
                        },
                        {
                            "id": "proj3-id",
                            "created_at": 1630519988.9949625,
                            "title": "Project 3",
                            "description": "Test description 3.",
                            "created_by": "user-id-3",
                            "started_at": 1630519986.9949625,
                            "updated_at": 1630515986.9949625,
                            "status": "ACTIVE",
                            "type": "SANDBOX",
                            "updated_by": "user-id-1",
                            "cohortList": {"cohorts": []},
                        },
                    ],
                    "pageInfo": {"endCursor": None},
                },
            }
        }

        projects = get_projects(client=self.mock_client)

        self.assertEqual(
            projects_expected,
            projects.to_list(),
        )

    def test_get_projects_pagination(self):
        """
        Test get projects for the initialized user, where the user has to
        page through projects.

        """
        project_1 = {
            "id": "proj1-id",
            "created_at": 1630515986.9949625,
            "updated_at": 1630515986.9949625,
            "title": "Project 1",
            "description": "Test description.",
            "created_by": "user-id-1",
            "started_at": 1630515986.9949625,
            "status": "ACTIVE",
            "type": "SANDBOX",
            "created_by": "user-id-1",
            "updated_by": "user-id-1",
            "cohorts": [],
        }
        project_2 = {
            "id": "proj2-id",
            "created_at": 1630517986.9949625,
            "updated_at": 1630515986.9949625,
            "title": "Project 2",
            "description": "Test description 2.",
            "created_by": "user-id-2",
            "started_at": 1630517986.9949625,
            "status": "ACTIVE",
            "type": "SANDBOX",
            "updated_by": "user-id-1",
            "cohorts": [],
        }
        project_3 = {
            "id": "proj3-id",
            "created_at": 1630519986.9949625,
            "updated_at": 1630515986.9949625,
            "title": "Project 3",
            "description": "Test description 3.",
            "created_by": "user-id-3",
            "started_at": 1630519986.9949625,
            "status": "ACTIVE",
            "type": "SANDBOX",
            "updated_by": "user-id-1",
            "cohorts": [],
        }

        self.mock_client.execute = mock.Mock()
        self.mock_client.execute.side_effect = [
            {
                "org": {
                    "id": "org-rune,org",
                    "projectList": {
                        "projects": [
                            {
                                "id": "proj1-id",
                                "created_at": 1630515986.9949625,
                                "updated_at": 1630515986.9949625,
                                "title": "Project 1",
                                "description": "Test description.",
                                "created_by": "user-id-1",
                                "started_at": 1630515986.9949625,
                                "status": "ACTIVE",
                                "type": "SANDBOX",
                                "created_by": "user-id-1",
                                "updated_by": "user-id-1",
                                "cohortList": {"cohorts": []},
                            },
                            {
                                "id": "proj2-id",
                                "created_at": 1630517986.9949625,
                                "updated_at": 1630515986.9949625,
                                "title": "Project 2",
                                "description": "Test description 2.",
                                "created_by": "user-id-2",
                                "started_at": 1630517986.9949625,
                                "status": "ACTIVE",
                                "type": "SANDBOX",
                                "updated_by": "user-id-1",
                                "cohortList": {"cohorts": []},
                            },
                        ],
                        "pageInfo": {"endCursor": "test_check_next"},
                    },
                }
            },
            {
                "org": {
                    "id": "org-rune,org",
                    "projectList": {
                        "projects": [
                            {
                                "id": "proj3-id",
                                "created_at": 1630519986.9949625,
                                "updated_at": 1630515986.9949625,
                                "title": "Project 3",
                                "description": "Test description 3.",
                                "created_by": "user-id-3",
                                "started_at": 1630519986.9949625,
                                "status": "ACTIVE",
                                "type": "SANDBOX",
                                "updated_by": "user-id-1",
                                "cohortList": {"cohorts": []},
                            }
                        ],
                        "pageInfo": {"endCursor": None},
                    },
                }
            },
        ]

        projects = get_projects(client=self.mock_client)

        self.assertEqual(
            [project_1, project_2, project_3],
            projects.to_list(),
        )

    def test_project_patient_attributes(self):
        """
        Test attributes for an initialized Project Patient.

        """
        example_metrics = MetricSet(
            [
                Metric(
                    type="LAST_UPLOAD",
                    data_type="APPLEWATCH_TREMOR",
                    value=1648235452.965804,
                    time_interval="FOURTEEN_DAYS",
                    created_at=1673467628.837126,
                    updated_at=1673467628.837126,
                    id="1bfcf11f-8d5c-4714-b774-6b47e03c5900",
                ),
            ]
        )
        test_project = ProjectPatientMetadata(
            id="proj-patient-id",
            created_at=1630515986.9949625,
            updated_at=1630515986.9949625,
            created_by="user-1",
            updated_by="user-1",
            started_at=1630515986.9949625,
            metrics=example_metrics,
        )

        self.assertEqual("proj-patient-id", test_project.id)
        self.assertEqual(1630515986.9949625, test_project.created_at)
        self.assertEqual(1630515986.9949625, test_project.updated_at)
        self.assertEqual("user-1", test_project.created_by)
        self.assertEqual("user-1", test_project.updated_by)
        self.assertEqual(example_metrics, test_project.metrics)

    def test_cohort_attributes(self):
        """
        Test attributes for an initialized Cohort.

        """
        test_cohort = Cohort(
            id="cohort1-id",
            created_at=1630515986.9949625,
            updated_at=1630515986.9949625,
            title="Cohort 1",
            description="Test description.",
            created_by="user-1",
            updated_by="user-1",
        )

        self.assertEqual("cohort1-id", test_cohort.id)
        self.assertEqual(1630515986.9949625, test_cohort.created_at)
        self.assertEqual("Cohort 1", test_cohort.title)
        self.assertEqual(1630515986.9949625, test_cohort.updated_at)
        self.assertEqual("user-1", test_cohort.created_by)
        self.assertEqual("user-1", test_cohort.updated_by)

    def test_cohort_patient_attributes(self):
        """
        Test attributes for an initialized Cohort Patient.

        """
        example_metrics = MetricSet(
            [
                Metric(
                    type="LAST_UPLOAD",
                    data_type="APPLEWATCH_TREMOR",
                    value=1648235452.965804,
                    time_interval="FOURTEEN_DAYS",
                    created_at=1673467628.837126,
                    updated_at=1673467628.837126,
                    id="1bfcf11f-8d5c-4714-b774-6b47e03c5900",
                ),
            ]
        )
        test_project = CohortPatientMetadata(
            id="cohort-patient-id",
            created_at=1630515986.9949625,
            updated_at=1630515986.9949625,
            created_by="user-1",
            updated_by="user-1",
            started_at=1630515986.9949625,
            metrics=example_metrics,
        )

        self.assertEqual("cohort-patient-id", test_project.id)
        self.assertEqual(1630515986.9949625, test_project.created_at)
        self.assertEqual(1630515986.9949625, test_project.updated_at)
        self.assertEqual("user-1", test_project.created_by)
        self.assertEqual("user-1", test_project.updated_by)
        self.assertEqual(example_metrics, test_project.metrics)

    def test_get_project_patients_basic(self):
        """
        Test get project patients for the initialized user.

        """
        self.mock_client.execute = mock.Mock()
        project_patients_expected = [
            {
                "id": "patient-1",
                "metrics": [
                    {
                        "id": "metric-1",
                        "updated_at": 1673467628.837126,
                        "created_at": 1673467628.837126,
                        "type": "LAST_UPLOAD",
                        "data_type": "APPLEWATCH_TREMOR",
                        "value": 1648235452.965804,
                        "time_interval": "FOURTEEN_DAYS",
                    },
                ],
                "code_name": "code name 1",
                "created_at": 1673467625.063822,
                "updated_at": 1673467625.063822,
                "created_by": "user 1",
                "updated_by": "user 2",
            },
            {
                "id": "patient-2",
                "metrics": [
                    {
                        "id": "metric-2",
                        "updated_at": 1673467628.837126,
                        "created_at": 1673467628.837126,
                        "type": "LAST_UPLOAD",
                        "data_type": "APPLEWATCH_TREMOR",
                        "value": 1648235452.965804,
                        "time_interval": "FOURTEEN_DAYS",
                    },
                ],
                "code_name": "code name 2",
                "created_at": 1673467625.063822,
                "updated_at": 1673467625.063822,
                "created_by": "user 2",
                "updated_by": "user 3",
            },
        ]

        self.mock_client.execute.return_value = {
            "project": {
                "projectPatientList": {
                    "projectPatients": [
                        {
                            "patient": {"id": "patient-1"},
                            "metricList": {
                                "metrics": [
                                    {
                                        "id": "metric-1",
                                        "updated_at": 1673467628.837126,
                                        "created_at": 1673467628.837126,
                                        "type": "LAST_UPLOAD",
                                        "data_type": "APPLEWATCH_TREMOR",
                                        "value": 1648235452.965804,
                                        "time_interval": "FOURTEEN_DAYS",
                                    },
                                ]
                            },
                            "code_name": "code name 1",
                            "created_at": 1673467625.063822,
                            "updated_at": 1673467625.063822,
                            "created_by": "user 1",
                            "updated_by": "user 2",
                        },
                        {
                            "patient": {"id": "patient-2"},
                            "metricList": {
                                "metrics": [
                                    {
                                        "id": "metric-2",
                                        "updated_at": 1673467628.837126,
                                        "created_at": 1673467628.837126,
                                        "type": "LAST_UPLOAD",
                                        "data_type": "APPLEWATCH_TREMOR",
                                        "value": 1648235452.965804,
                                        "time_interval": "FOURTEEN_DAYS",
                                    },
                                ]
                            },
                            "code_name": "code name 2",
                            "created_at": 1673467625.063822,
                            "updated_at": 1673467625.063822,
                            "created_by": "user 2",
                            "updated_by": "user 3",
                        },
                    ],
                    "pageInfo": {"codeNameEndCursor": None},
                }
            }
        }

        project_patients = get_project_patients(
            client=self.mock_client, project_id="test-project-1"
        )

        self.assertEqual(
            project_patients_expected,
            project_patients.to_list(),
        )

    def test_get_project_patients_pagination(self):
        """
        Test get project patients pagination for the initialized user.

        """
        self.mock_client.execute = mock.Mock()
        project_patients_expected = [
            {
                "id": "patient-1",
                "metrics": [
                    {
                        "id": "metric-1",
                        "updated_at": 1673467628.837126,
                        "created_at": 1673467628.837126,
                        "type": "LAST_UPLOAD",
                        "data_type": "APPLEWATCH_TREMOR",
                        "value": 1648235452.965804,
                        "time_interval": "FOURTEEN_DAYS",
                    },
                ],
                "code_name": "code name 1",
                "created_at": 1673467625.063822,
                "updated_at": 1673467625.063822,
                "created_by": "user 1",
                "updated_by": "user 2",
            },
            {
                "id": "patient-2",
                "metrics": [
                    {
                        "id": "metric-2",
                        "updated_at": 1673467628.837126,
                        "created_at": 1673467628.837126,
                        "type": "LAST_UPLOAD",
                        "data_type": "APPLEWATCH_TREMOR",
                        "value": 1648235452.965804,
                        "time_interval": "FOURTEEN_DAYS",
                    },
                ],
                "code_name": "code name 2",
                "created_at": 1673467625.063822,
                "updated_at": 1673467625.063822,
                "created_by": "user 2",
                "updated_by": "user 3",
            },
        ]

        self.mock_client.execute.side_effect = [
            {
                "project": {
                    "projectPatientList": {
                        "projectPatients": [
                            {
                                "patient": {"id": "patient-1"},
                                "metricList": {
                                    "metrics": [
                                        {
                                            "id": "metric-1",
                                            "updated_at": 1673467628.837126,
                                            "created_at": 1673467628.837126,
                                            "type": "LAST_UPLOAD",
                                            "data_type": "APPLEWATCH_TREMOR",
                                            "value": 1648235452.965804,
                                            "time_interval": "FOURTEEN_DAYS",
                                        },
                                    ]
                                },
                                "code_name": "code name 1",
                                "created_at": 1673467625.063822,
                                "updated_at": 1673467625.063822,
                                "created_by": "user 1",
                                "updated_by": "user 2",
                            },
                        ],
                        "pageInfo": {"codeNameEndCursor": "code name 1"},
                    }
                }
            },
            {
                "project": {
                    "projectPatientList": {
                        "projectPatients": [
                            {
                                "patient": {"id": "patient-2"},
                                "metricList": {
                                    "metrics": [
                                        {
                                            "id": "metric-2",
                                            "updated_at": 1673467628.837126,
                                            "created_at": 1673467628.837126,
                                            "type": "LAST_UPLOAD",
                                            "data_type": "APPLEWATCH_TREMOR",
                                            "value": 1648235452.965804,
                                            "time_interval": "FOURTEEN_DAYS",
                                        },
                                    ]
                                },
                                "code_name": "code name 2",
                                "created_at": 1673467625.063822,
                                "updated_at": 1673467625.063822,
                                "created_by": "user 2",
                                "updated_by": "user 3",
                            },
                        ],
                        "pageInfo": {"codeNameEndCursor": None},
                    }
                }
            },
        ]

        project_patients = get_project_patients(
            client=self.mock_client, project_id="test-project-1"
        )

        self.assertEqual(
            project_patients_expected,
            project_patients.to_list(),
        )

    def test_get_cohort_patients_basic(self):
        """
        Test get cohort patients for the initialized user.

        """
        self.mock_client.execute = mock.Mock()
        cohort_patients_expected = [
            {
                "id": "patient-1",
                "metrics": [
                    {
                        "id": "metric-1",
                        "updated_at": 1673467628.837126,
                        "created_at": 1673467628.837126,
                        "type": "LAST_UPLOAD",
                        "data_type": "APPLEWATCH_TREMOR",
                        "value": 1648235452.965804,
                        "time_interval": "FOURTEEN_DAYS",
                    },
                ],
                "code_name": "code name 1",
                "created_at": 1673467625.063822,
                "updated_at": 1673467625.063822,
                "created_by": "user 1",
                "updated_by": "user 2",
            },
            {
                "id": "patient-2",
                "metrics": [
                    {
                        "id": "metric-2",
                        "updated_at": 1673467628.837126,
                        "created_at": 1673467628.837126,
                        "type": "LAST_UPLOAD",
                        "data_type": "APPLEWATCH_TREMOR",
                        "value": 1648235452.965804,
                        "time_interval": "FOURTEEN_DAYS",
                    },
                ],
                "code_name": "code name 2",
                "created_at": 1673467625.063822,
                "updated_at": 1673467625.063822,
                "created_by": "user 2",
                "updated_by": "user 3",
            },
        ]

        self.mock_client.execute.return_value = {
            "cohort": {
                "cohortPatientList": {
                    "cohortPatients": [
                        {
                            "patient": {"id": "patient-1"},
                            "metricList": {
                                "metrics": [
                                    {
                                        "id": "metric-1",
                                        "updated_at": 1673467628.837126,
                                        "created_at": 1673467628.837126,
                                        "type": "LAST_UPLOAD",
                                        "data_type": "APPLEWATCH_TREMOR",
                                        "value": 1648235452.965804,
                                        "time_interval": "FOURTEEN_DAYS",
                                    },
                                ]
                            },
                            "code_name": "code name 1",
                            "created_at": 1673467625.063822,
                            "updated_at": 1673467625.063822,
                            "created_by": "user 1",
                            "updated_by": "user 2",
                        },
                        {
                            "patient": {"id": "patient-2"},
                            "metricList": {
                                "metrics": [
                                    {
                                        "id": "metric-2",
                                        "updated_at": 1673467628.837126,
                                        "created_at": 1673467628.837126,
                                        "type": "LAST_UPLOAD",
                                        "data_type": "APPLEWATCH_TREMOR",
                                        "value": 1648235452.965804,
                                        "time_interval": "FOURTEEN_DAYS",
                                    },
                                ]
                            },
                            "code_name": "code name 2",
                            "created_at": 1673467625.063822,
                            "updated_at": 1673467625.063822,
                            "created_by": "user 2",
                            "updated_by": "user 3",
                        },
                    ],
                    "pageInfo": {"codeNameEndCursor": None},
                }
            }
        }

        cohort_patients = get_cohort_patients(
            client=self.mock_client, cohort_id="test-cohort-1"
        )

        self.assertEqual(
            cohort_patients_expected,
            cohort_patients.to_list(),
        )

    def test_get_cohort_patients_pagination(self):
        """
        Test get cohort patients pagination for the initialized user.

        """
        self.mock_client.execute = mock.Mock()
        cohort_patients_expected = [
            {
                "id": "patient-1",
                "metrics": [
                    {
                        "id": "metric-1",
                        "updated_at": 1673467628.837126,
                        "created_at": 1673467628.837126,
                        "type": "LAST_UPLOAD",
                        "data_type": "APPLEWATCH_TREMOR",
                        "value": 1648235452.965804,
                        "time_interval": "FOURTEEN_DAYS",
                    },
                ],
                "code_name": "code name 1",
                "created_at": 1673467625.063822,
                "updated_at": 1673467625.063822,
                "created_by": "user 1",
                "updated_by": "user 2",
            },
            {
                "id": "patient-2",
                "metrics": [
                    {
                        "id": "metric-2",
                        "updated_at": 1673467628.837126,
                        "created_at": 1673467628.837126,
                        "type": "LAST_UPLOAD",
                        "data_type": "APPLEWATCH_TREMOR",
                        "value": 1648235452.965804,
                        "time_interval": "FOURTEEN_DAYS",
                    },
                ],
                "code_name": "code name 2",
                "created_at": 1673467625.063822,
                "updated_at": 1673467625.063822,
                "created_by": "user 2",
                "updated_by": "user 3",
            },
        ]

        self.mock_client.execute.side_effect = [
            {
                "cohort": {
                    "cohortPatientList": {
                        "cohortPatients": [
                            {
                                "patient": {"id": "patient-1"},
                                "metricList": {
                                    "metrics": [
                                        {
                                            "id": "metric-1",
                                            "updated_at": 1673467628.837126,
                                            "created_at": 1673467628.837126,
                                            "type": "LAST_UPLOAD",
                                            "data_type": "APPLEWATCH_TREMOR",
                                            "value": 1648235452.965804,
                                            "time_interval": "FOURTEEN_DAYS",
                                        },
                                    ]
                                },
                                "code_name": "code name 1",
                                "created_at": 1673467625.063822,
                                "updated_at": 1673467625.063822,
                                "created_by": "user 1",
                                "updated_by": "user 2",
                            },
                        ],
                        "pageInfo": {"codeNameEndCursor": "code name 1"},
                    }
                }
            },
            {
                "cohort": {
                    "cohortPatientList": {
                        "cohortPatients": [
                            {
                                "patient": {"id": "patient-2"},
                                "metricList": {
                                    "metrics": [
                                        {
                                            "id": "metric-2",
                                            "updated_at": 1673467628.837126,
                                            "created_at": 1673467628.837126,
                                            "type": "LAST_UPLOAD",
                                            "data_type": "APPLEWATCH_TREMOR",
                                            "value": 1648235452.965804,
                                            "time_interval": "FOURTEEN_DAYS",
                                        },
                                    ]
                                },
                                "code_name": "code name 2",
                                "created_at": 1673467625.063822,
                                "updated_at": 1673467625.063822,
                                "created_by": "user 2",
                                "updated_by": "user 3",
                            },
                        ],
                        "pageInfo": {"codeNameEndCursor": None},
                    }
                }
            },
        ]

        cohort_patients = get_cohort_patients(
            client=self.mock_client, cohort_id="test-cohort-1"
        )

        self.assertEqual(
            cohort_patients_expected,
            cohort_patients.to_list(),
        )



"""
Fetch metadata about Rune platform users.

"""

from typing import Optional

from .client import GraphClient, global_graph_client
from .common import ItemBase


class User(ItemBase):
    """
    A user of the Rune platform.

    """

    def __init__(
        self,
        id: str,
        name: str,
        created_at: float,
        active_org_id: str,
        active_org_name: str,
        **attributes
    ):
        """
        Initialize with metadata.

        Args:
            id: User ID
            name: Human-readable display name for the user
            created_at: When the user was created (unix timestamp)
            active_org_id: ID of the user's currently active organization
            active_org_name: Display name of the user's currently
                active organization

        """
        norm_id = id
        self.name = name
        self.created_at = created_at
        self.active_org_id = active_org_id
        self.active_org_name = active_org_name

        super().__init__(
            id=norm_id,
            name=name,
            created_at=created_at,
            active_org_id=active_org_id,
            active_org_name=active_org_name,
            **attributes,
        )

    @staticmethod
    def normalize_id(user_id: str) -> str:
        """
        Strip resource prefix and suffix from the user ID (if they exist).

        Args:
            user_id: User ID

        """
        if user_id.startswith("user-"):
            user_id = user_id[5:]

        if user_id.endswith(",user"):
            user_id = user_id[:-5]

        return user_id


def get_current_user(client: Optional[GraphClient] = None) -> User:
    """
    Get information about the current user (based on the API credentials).

    Args:
        client: If specified, this client is used to fetch metadata from the
            API. Otherwise, the global GraphClient is used.
    """
    client = client or global_graph_client()
    query = """
        query getUser {
            user {
                id
                created_at: created
                name: displayName
                defaultMembership {
                    org {
                        id
                        name: displayName
                    }
                }
                email
            }
        }
    """

    result = client.execute(statement=query)

    user_attrs = result["user"]

    org = user_attrs.get("defaultMembership", {}).get("org")

    return User(
        id=user_attrs["id"],
        name=user_attrs.get("name"),
        created_at=user_attrs.get("created_at"),
        active_org_id=org.get("id"),
        active_org_name=org.get("name"),
        email=user_attrs.get("email"),
    )



"""
Tests for fetching user metadata.

"""
from unittest import TestCase, mock

from runeq.config import Config
from runeq.resources.client import GraphClient
from runeq.resources.user import User, get_current_user


class TestUser(TestCase):
    """
    Unit tests for the User class.

    """

    def setUp(self):
        """
        Set up mock graph client for testing.

        """
        self.mock_client = GraphClient(
            Config(client_key_id="test", client_access_key="config")
        )

    def test_attributes(self):
        """
        Test attributes for an initialized User.

        """
        test_user = User(
            id="user1-id",
            created_at=1629300943.9179766,
            name="user1",
            active_org_id="org1-id",
            active_org_name="org1",
        )

        self.assertEqual("user1-id", test_user.id)
        self.assertEqual(1629300943.9179766, test_user.created_at)
        self.assertEqual("user1", test_user.name)
        self.assertEqual("org1-id", test_user.active_org_id)
        self.assertEqual("org1", test_user.active_org_name)

    def test_repr(self):
        """
        Test __repr__

        """
        test_user = User(
            id="user1-id",
            created_at=1629300943.9179766,
            name="User 1",
            active_org_id="org1-id",
            active_org_name="Org 1",
        )

        self.assertEqual('User(id="user1-id", name="User 1")', repr(test_user))

    def test_normalize_id(self):
        """
        Test strip id of resource prefix and suffix if they exist.

        """
        self.assertEqual(
            "d4b1c627bd464fe0a5ed940cc8e8e485",
            User.normalize_id("user-d4b1c627bd464fe0a5ed940cc8e8e485,user"),
        )

        self.assertEqual(
            "d4b1c627bd464fe0a5ed940cc8e8e485",
            User.normalize_id("d4b1c627bd464fe0a5ed940cc8e8e485,user"),
        )

        self.assertEqual(
            "d4b1c627bd464fe0a5ed940cc8e8e485",
            User.normalize_id("user-d4b1c627bd464fe0a5ed940cc8e8e485"),
        )

        self.assertEqual(
            "d4b1c627bd464fe0a5ed940cc8e8e485",
            User.normalize_id("d4b1c627bd464fe0a5ed940cc8e8e485"),
        )

    def test_get_user(self):
        """
        Test get org for specified org_id.

        """
        self.mock_client.execute = mock.Mock()
        self.mock_client.execute.side_effect = [
            {
                "user": {
                    "id": "user1-id",
                    "created_at": 1629300943.9179766,
                    "name": "user1",
                    "defaultMembership": {"org": {"id": "org1-id", "name": "org1"}},
                    "email": "user1@email.com",
                }
            }
        ]

        user = get_current_user(client=self.mock_client)

        self.assertEqual(
            {
                "id": "user1-id",
                "created_at": 1629300943.9179766,
                "name": "user1",
                "active_org_id": "org1-id",
                "active_org_name": "org1",
                "email": "user1@email.com",
            },
            user.to_dict(),
        )



"""
Tests for fetching stream data.

"""
from datetime import datetime, timezone
from unittest import TestCase, mock

from runeq.config import Config
from runeq.resources.client import StreamClient
from runeq.resources.stream import get_stream_availability, get_stream_data


class TestStreamData(TestCase):
    """
    Unit tests for the data queries in the Stream class.

    """

    def setUp(self):
        """
        Set up mock graph client for testing.

        """
        self.stream_client = StreamClient(
            Config(client_key_id="test", client_access_key="config")
        )
        self.maxDiff = None

    @mock.patch("runeq.resources.client.requests")
    def test_get_stream_data_csv(self, mock_requests):
        """
        Test get a stream as csv with specific stream_id and optional
        parameters.

        """
        expected_data = """time,acceleration,measurement_duration_ns
2022-07-28T14:26:45.167568Z,0.020525138825178146,20000000
2022-07-28T14:26:45.361596Z,0.020834974944591522,20000000
2022-07-28T14:26:45.361796Z,0.021182861179113388,20000000
2022-07-28T14:26:45.3618588Z,0.022172993049025536,20000000
2022-07-28T14:26:45.3620749Z,0.02356025017797947,20000000
2022-07-28T14:26:45.362149Z,0.024860087782144547,20000000
2022-07-28T14:26:45.36221Z,0.026072751730680466,20000000"""

        # Mock a paginated response
        mock_response1 = mock.Mock()
        mock_response1.ok = True
        mock_response1.headers = {"X-Rune-Next-Page-Token": "foobar"}
        mock_response1.text = expected_data

        mock_response2 = mock.Mock()
        mock_response2.ok = True
        mock_response2.headers = {}
        mock_response2.text = expected_data

        mock_requests.get.side_effect = [mock_response1, mock_response2]

        stream = get_stream_data(
            "test_stream_id",
            client=self.stream_client,
        )

        expected = [expected_data, expected_data]
        actual = list(stream)
        self.assertEqual(expected, actual)
        self.assertEqual(mock_requests.get.call_count, 2)

    @mock.patch("runeq.resources.client.requests")
    def test_get_stream_data_params(self, mock_requests):
        """
        Check the request construction for fetching stream data

        """
        # Mock an empty response - this test is just for the request
        mock_response = mock.Mock()
        mock_response.ok = True
        mock_response.headers = {}
        mock_response.json.return_value = {}
        mock_requests.get.return_value = mock_response

        data = get_stream_data(
            "stream1",
            start_time=1691760000,
            end_time=1691762094.123,
            format="json",
            limit=10,
            page_token="page-token",
            timestamp="unix",
            timezone_name="America/Los_Angeles",
            translate_enums=False,
            client=self.stream_client,
        )
        # call list() to consume the iterator
        _ = list(data)

        expected_headers = {
            "X-Rune-Client-Key-ID": "test",
            "X-Rune-Client-Access-Key": "config",
        }

        mock_requests.get.assert_called_once_with(
            "https://stream.runelabs.io/v2/streams/stream1",
            headers=expected_headers,
            params={
                "start_time": 1691760000,
                "start_time_ns": None,
                "end_time": 1691762094.123,
                "end_time_ns": None,
                "format": "json",
                "limit": 10,
                "page_token": "page-token",
                "timestamp": "unix",
                "timezone": None,
                "timezone_name": "America/Los_Angeles",
                "translate_enums": False,
            },
        )

        mock_requests.get.reset_mock()

        # Test timestamp conversion
        data = get_stream_data(
            "stream2",
            start_time=datetime(2023, 8, 1, tzinfo=timezone.utc),
            end_time=datetime(2023, 8, 11, 10, 26, 45, 1, tzinfo=timezone.utc),
            client=self.stream_client,
        )
        # call list() to consume the iterator
        _ = list(data)

        mock_requests.get.assert_called_once_with(
            "https://stream.runelabs.io/v2/streams/stream2",
            headers=expected_headers,
            params={
                "start_time": 1690848000.0,
                "start_time_ns": None,
                "end_time": 1691749605.000001,
                "end_time_ns": None,
                "format": "csv",
                "limit": None,
                "page_token": None,
                "timestamp": "iso",
                "timezone": None,
                "timezone_name": None,
                "translate_enums": True,
            },
        )

    @mock.patch("runeq.resources.client.requests")
    def test_get_stream_data_json(self, mock_requests):
        """
        Test get a stream as json with specific stream_id and optional
        parameters.

        """
        expected_data = {
            "cardinality": 10,
            "data": {
                "acceleration": [
                    0.020525138825178146,
                    0.020834974944591522,
                    0.021182861179113388,
                    0.022172993049025536,
                    0.02356025017797947,
                    0.024860087782144547,
                    0.026072751730680466,
                    0.027338741347193718,
                    0.028795091435313225,
                    0.029819512739777565,
                ],
                "measurement_duration_ns": [
                    20000000,
                    20000000,
                    20000000,
                    20000000,
                    20000000,
                    20000000,
                    20000000,
                    20000000,
                    20000000,
                    20000000,
                ],
                "time": [
                    "2022-07-28T14:26:45.167568Z",
                    "2022-07-28T14:26:45.361596Z",
                    "2022-07-28T14:26:45.361796Z",
                    "2022-07-28T14:26:45.3618588Z",
                    "2022-07-28T14:26:45.3620749Z",
                    "2022-07-28T14:26:45.362149Z",
                    "2022-07-28T14:26:45.36221Z",
                    "2022-07-28T14:26:45.362269Z",
                    "2022-07-28T14:26:45.36234Z",
                    "2022-07-28T14:26:45.3624Z",
                ],
            },
        }

        # Mock a paginated response
        mock_response1 = mock.Mock()
        mock_response1.ok = True
        mock_response1.headers = {"X-Rune-Next-Page-Token": "foobar"}
        mock_response1.json.return_value = expected_data

        mock_response2 = mock.Mock()
        mock_response2.ok = True
        mock_response2.headers = {}
        mock_response2.json.return_value = expected_data

        mock_requests.get.side_effect = [mock_response1, mock_response2]

        stream = get_stream_data(
            "test_stream_id",
            format="json",
            client=self.stream_client,
        )

        expected = [expected_data, expected_data]
        actual = list(stream)
        self.assertEqual(expected, actual)

    @mock.patch("runeq.resources.client.requests")
    def test_get_stream_availability_csv(self, mock_requests):
        """
        Test get a stream availability as csv for a specific stream_id.

        """
        expected_data = """time,availability
2022-07-28T14:26:45.167568Z,1
2022-07-28T14:26:45.361596Z,1
2022-07-28T14:26:45.361796Z,0
2022-07-28T14:26:45.3618588Z,0
2022-07-28T14:26:45.3620749Z,0
2022-07-28T14:26:45.362149Z,1
2022-07-28T14:26:45.36221Z,1
"""

        # Mock a paginated response
        mock_response1 = mock.Mock()
        mock_response1.ok = True
        mock_response1.headers = {"X-Rune-Next-Page-Token": "foobar"}
        mock_response1.text = expected_data

        mock_response2 = mock.Mock()
        mock_response2.ok = True
        mock_response2.headers = {}
        mock_response2.text = expected_data

        mock_requests.get.side_effect = [mock_response1, mock_response2]

        availability = get_stream_availability(
            stream_ids="test_stream_id",
            start_time=123,
            end_time=345,
            resolution=300,
            format="csv",
            client=self.stream_client,
        )

        expected = [expected_data, expected_data]
        actual = list(availability)
        self.assertEqual(expected, actual)

    @mock.patch("runeq.resources.client.requests")
    def test_get_stream_availability_json(self, mock_requests):
        """
        Test get a stream availability as json for a specific stream_id.

        """
        expected_data = {
            "data": {
                "time": [
                    "2022-07-28T14:25:00Z",
                    "2022-07-28T14:30:00Z",
                    "2022-07-28T14:35:00Z",
                    "2022-07-28T14:40:00Z",
                    "2022-07-28T14:45:00Z",
                    "2022-07-28T14:50:00Z",
                    "2022-07-28T14:55:00Z",
                    "2022-07-28T15:00:00Z",
                    "2022-07-28T15:05:00Z",
                    "2022-07-28T15:10:00Z",
                ],
                "availability": [1, 1, 0, 0, 1, 1, 0, 1, 1, 1],
            },
            "approx_available_duration_s": 3000,
            "cardinality": 10,
        }

        # Mock a paginated response
        mock_response1 = mock.Mock()
        mock_response1.ok = True
        mock_response1.headers = {"X-Rune-Next-Page-Token": "foobar"}
        mock_response1.json.return_value = expected_data

        mock_response2 = mock.Mock()
        mock_response2.ok = True
        mock_response2.headers = {}
        mock_response2.json.return_value = expected_data

        mock_requests.get.side_effect = [mock_response1, mock_response2]

        availability = get_stream_availability(
            stream_ids="test_stream_id",
            start_time=123,
            end_time=345,
            resolution=300,
            format="json",
            client=self.stream_client,
        )

        expected = [expected_data, expected_data]
        actual = list(availability)
        self.assertEqual(expected, actual)

    @mock.patch("runeq.resources.client.requests")
    def test_get_batch_stream_availability(self, mock_requests):
        """
        Test get batch stream availability for multiple stream_ids.

        """
        expected_data = """time,availability
2022-07-28T14:26:45.167568Z,1
2022-07-28T14:26:45.361596Z,1
2022-07-28T14:26:45.361796Z,0
2022-07-28T14:26:45.3618588Z,0
2022-07-28T14:26:45.3620749Z,0
2022-07-28T14:26:45.362149Z,1
2022-07-28T14:26:45.36221Z,1
"""

        # Mock a paginated response
        mock_response1 = mock.Mock()
        mock_response1.ok = True
        mock_response1.headers = {"X-Rune-Next-Page-Token": "foobar"}
        mock_response1.text = expected_data

        mock_response2 = mock.Mock()
        mock_response2.ok = True
        mock_response2.headers = {}
        mock_response2.text = expected_data

        mock_requests.get.side_effect = [mock_response1, mock_response2]

        # Must include batch_operation when querying >1 stream
        with self.assertRaisesRegex(ValueError, "batch_operation must be specified"):
            get_stream_availability(
                stream_ids=["test_stream_id1", "test_stream_id2"],
                start_time=123,
                end_time=345,
                resolution=300,
                format="csv",
                client=self.stream_client,
            ).__next__()

        availability = get_stream_availability(
            stream_ids=["test_stream_id1", "test_stream_id2"],
            start_time=123,
            end_time=345,
            resolution=300,
            batch_operation="all",
            client=self.stream_client,
        )

        expected = [expected_data, expected_data]
        actual = list(availability)
        self.assertEqual(expected, actual)



"""
Tests for the V2 SDK Client.

"""

from unittest import TestCase, mock

from runeq import errors
from runeq.config import BaseConfig
from runeq.resources.client import (
    GraphClient,
    StreamClient,
    global_graph_client,
    global_stream_client,
    initialize,
    initialize_with_config,
)
from runeq.resources.stream import get_stream_data


class TestInitialize(TestCase):
    """
    Unit tests for the initialization of the user's credentials.

    """

    def test_init_invalid(self):
        """
        Test that access is restricted with no initialization.

        """
        with self.assertRaises(errors.InitializationError):
            get_stream_data("stream_id").__next__()

    def test_init_with_access_keys(self):
        """
        Test that the user is initialized but with invalid access keys.

        """
        initialize(
            access_token_id="foo",
            access_token_secret="bar",
        )

        with self.assertRaisesRegex(errors.APIError, "401 InvalidAuthentication"):
            get_stream_data("stream_id").__next__()

    def test_init_with_client_keys(self):
        """
        Test that the user is initialized but with invalid client keys.

        """
        initialize(
            client_key_id="foo",
            client_access_key="bar",
        )

        with self.assertRaisesRegex(errors.APIError, "401 InvalidAuthentication"):
            get_stream_data("stream_id").__next__()

    def test_initialize_with_config(self):
        """
        Test initializing with a config object.

        """
        config = mock.Mock(spec=BaseConfig)
        config.graph_url = "graph-url"
        config.stream_url = "stream-url"
        config.auth_headers = {"hello": "world"}

        initialize_with_config(config)

        graph_client = global_graph_client()
        stream_client = global_stream_client()

        self.assertIs(graph_client.config, config)
        self.assertIs(stream_client.config, config)


class TestStreamClient(TestCase):
    """
    Unit tests for StreamClient

    """

    def _setup_mock_response(self, status_code, json_body):
        """Set up a mock response"""
        mock_response = mock.Mock()
        mock_response.ok = False
        mock_response.status_code = status_code
        mock_response.json.return_value = json_body
        return mock_response

    @mock.patch("runeq.resources.client.requests.get")
    def test_refresh_auth_on_4xx(self, mock_get):
        """If a request fails with a 4xx status code, refresh auth and retry"""
        config = mock.Mock(spec=BaseConfig)
        config.stream_url = ""
        config.refresh_auth.return_value = True

        stream_client = StreamClient(config)

        mock_response = self._setup_mock_response(400, {"error": "auth error"})
        mock_get.return_value = mock_response

        with self.assertRaises(errors.APIError):
            list(stream_client.get_data("/v2/streams/123"))

        self.assertEqual(mock_get.call_count, 2)
        config.refresh_auth.assert_called_once()

    @mock.patch("runeq.resources.client.requests.get")
    def test_no_retry_on_5xx(self, mock_get):
        """If a request fails with a 5xx status code, do not retry"""
        config = mock.Mock(spec=BaseConfig)
        config.stream_url = ""
        config.refresh_auth.return_value = True

        stream_client = StreamClient(config)

        mock_response = self._setup_mock_response(500, {"error": "internal"})
        mock_get.return_value = mock_response

        with self.assertRaises(errors.APIError):
            list(stream_client.get_data("/v2/streams/123"))

        self.assertEqual(mock_get.call_count, 1)
        config.refresh_auth.assert_not_called()


class TestGraphClient(TestCase):
    """
    Unit tests for GraphClient

    """

    @mock.patch("runeq.resources.client.GQLClient")
    def test_refresh_auth_on_error(self, mock_client_cls):
        """If a request fails with any error, refresh auth and retry"""
        config = mock.Mock(spec=BaseConfig)
        config.graph_url = ""
        config.auth_headers = {"hello": "world"}

        graph_client = GraphClient(config)

        mock_execute = mock_client_cls().execute
        excecute_err = ValueError("test exception")
        mock_execute.side_effect = excecute_err

        with self.assertRaisesRegex(ValueError, "test exception"):
            graph_client.execute("query fakeQuery($input: Input) { id }")

        self.assertEqual(mock_execute.call_count, 2)
        config.refresh_auth.assert_called_once()



"""
Tests for fetching org metadata.

"""
from unittest import TestCase, mock

from runeq.config import Config
from runeq.resources.client import GraphClient
from runeq.resources.org import Org, get_org, get_orgs, set_active_org


class TestOrg(TestCase):
    """
    Unit tests for the Org class.

    """

    def setUp(self):
        """
        Set up mock graph client for testing.

        """
        self.mock_client = GraphClient(
            Config(client_key_id="test", client_access_key="config")
        )

    def test_attributes(self):
        """
        Test attributes for an initialized Org.

        """
        test_org = Org(
            id="org1-id",
            created_at=1629300943.9179766,
            name="org1",
            tags=("tag1", "tag2"),
        )

        self.assertEqual("org1-id", test_org.id)
        self.assertEqual(1629300943.9179766, test_org.created_at)
        self.assertEqual("org1", test_org.name)
        self.assertEqual(["tag1", "tag2"], test_org.tags)

    def test_repr(self):
        """
        Test __repr__

        """
        test_org = Org(
            id="org1-id",
            created_at=1629300943.9179766,
            name="Org 1",
        )
        self.assertEqual('Org(id="org1-id", name="Org 1")', repr(test_org))

    def test_normalize_id(self):
        """
        Test strip id of resource prefix and suffix if they exist.

        """
        self.assertEqual(
            "d4b1c627bd464fe0a5ed940cc8e8e485",
            Org.normalize_id("org-d4b1c627bd464fe0a5ed940cc8e8e485,org"),
        )

        self.assertEqual(
            "d4b1c627bd464fe0a5ed940cc8e8e485",
            Org.normalize_id("d4b1c627bd464fe0a5ed940cc8e8e485,org"),
        )

        self.assertEqual(
            "d4b1c627bd464fe0a5ed940cc8e8e485",
            Org.normalize_id("org-d4b1c627bd464fe0a5ed940cc8e8e485"),
        )

        self.assertEqual(
            "d4b1c627bd464fe0a5ed940cc8e8e485",
            Org.normalize_id("d4b1c627bd464fe0a5ed940cc8e8e485"),
        )

    def test_denormalize_id(self):
        """
        Test add id resource prefix and suffix if they don't exist.

        """
        self.assertEqual(
            "org-d4b1c627bd464fe0a5ed940cc8e8e485,org",
            Org.denormalize_id("org-d4b1c627bd464fe0a5ed940cc8e8e485,org"),
        )

        self.assertEqual(
            "org-d4b1c627bd464fe0a5ed940cc8e8e485,org",
            Org.denormalize_id("d4b1c627bd464fe0a5ed940cc8e8e485,org"),
        )

        self.assertEqual(
            "org-d4b1c627bd464fe0a5ed940cc8e8e485,org",
            Org.denormalize_id("org-d4b1c627bd464fe0a5ed940cc8e8e485"),
        )

        self.assertEqual(
            "org-d4b1c627bd464fe0a5ed940cc8e8e485,org",
            Org.denormalize_id("d4b1c627bd464fe0a5ed940cc8e8e485"),
        )

    def test_get_org(self):
        """
        Test get org for specified org_id.

        """
        self.mock_client.execute = mock.Mock(
            return_value={
                "user": {
                    "membershipList": {
                        "pageInfo": {"endCursor": None},
                        "memberships": [
                            {
                                "org": {
                                    "id": "org1-id",
                                    "created_at": 1571267538.391721,
                                    "name": "org1",
                                    "tags": [],
                                }
                            },
                            {
                                "org": {
                                    "id": "org2-id",
                                    "created_at": 1630515986.9949625,
                                    "name": "org2",
                                    "tags": ["tag1", "tag2"],
                                }
                            },
                        ],
                    }
                }
            }
        )

        org = get_org("org1-id", client=self.mock_client)
        self.assertEqual(
            {
                "id": "org1-id",
                "created_at": 1571267538.391721,
                "name": "org1",
                "tags": [],
            },
            org.to_dict(),
        )

        org = get_org("org2-id", client=self.mock_client)
        self.assertEqual(
            {
                "id": "org2-id",
                "created_at": 1630515986.9949625,
                "name": "org2",
                "tags": ["tag1", "tag2"],
            },
            org.to_dict(),
        )

    def test_get_orgs_basic(self):
        """
        Test get orgs for the initialized user.

        """
        self.mock_client.execute = mock.Mock()
        self.mock_client.execute.side_effect = [
            {
                "user": {
                    "membershipList": {
                        "pageInfo": {"endCursor": None},
                        "memberships": [
                            {
                                "org": {
                                    "id": "org1-id",
                                    "created_at": 1571267538.391721,
                                    "name": "org1",
                                    "tags": ["tag1"],
                                }
                            },
                            {
                                "org": {
                                    "id": "org2-id",
                                    "created_at": 1630515986.9949625,
                                    "name": "org2",
                                    "tags": ["tag2"],
                                }
                            },
                            {
                                "org": {
                                    "id": "org3-id",
                                    "created_at": 1649888079.066764,
                                    "name": "org3",
                                    "tags": [],
                                }
                            },
                        ],
                    }
                }
            }
        ]

        orgs = get_orgs(client=self.mock_client)

        self.assertEqual(
            [
                {
                    "created_at": 1571267538.391721,
                    "name": "org1",
                    "id": "org1-id",
                    "tags": ["tag1"],
                },
                {
                    "created_at": 1630515986.9949625,
                    "name": "org2",
                    "id": "org2-id",
                    "tags": ["tag2"],
                },
                {
                    "created_at": 1649888079.066764,
                    "name": "org3",
                    "id": "org3-id",
                    "tags": [],
                },
            ],
            orgs.to_list(),
        )

    def test_get_orgs_pagination(self):
        """
        Test get orgs for the initialized user, where the user has to
        page through orgs.

        """
        self.mock_client.execute = mock.Mock()
        self.mock_client.execute.side_effect = [
            {
                "user": {
                    "membershipList": {
                        "pageInfo": {"endCursor": "test_check_next"},
                        "memberships": [
                            {
                                "org": {
                                    "id": "org1-id",
                                    "created_at": 1571267538.391721,
                                    "name": "org1",
                                    "tags": [],
                                }
                            }
                        ],
                    }
                }
            },
            {
                "user": {
                    "membershipList": {
                        "pageInfo": {"endCursor": None},
                        "memberships": [
                            {
                                "org": {
                                    "id": "org2-id",
                                    "created_at": 1630515986.9949625,
                                    "name": "org2",
                                    "tags": [],
                                }
                            }
                        ],
                    }
                }
            },
        ]

        orgs = get_orgs(client=self.mock_client)

        self.assertEqual(
            [
                {
                    "created_at": 1571267538.391721,
                    "name": "org1",
                    "id": "org1-id",
                    "tags": [],
                },
                {
                    "created_at": 1630515986.9949625,
                    "name": "org2",
                    "id": "org2-id",
                    "tags": [],
                },
            ],
            orgs.to_list(),
        )

    def test_set_active_org(self):
        """
        Test setting the user's active organization

        """
        org_id = "org1-id"
        self.mock_client.execute = mock.Mock()
        self.mock_client.execute.side_effect = [
            {
                "updateDefaultMembership": {
                    "user": {
                        "defaultMembership": {
                            "org": {
                                "id": org_id,
                                "name": "org1",
                                "created_at": 1571267538.391721,
                                "tags": ["tag1"],
                            }
                        }
                    }
                }
            }
        ]

        new_org = set_active_org(org_id, self.mock_client)
        self.mock_client.execute.assert_called_once()
        self.assertEqual(
            new_org.to_dict(),
            {
                "created_at": 1571267538.391721,
                "name": "org1",
                "id": "org1-id",
                "tags": ["tag1"],
            },
        )



"""
Fetch data from Rune APIs.

By default, globally-initialized clients are used for all API requests (see
:class:`~runeq.resources.client.initialize`). Functions that make API requests
also accept optional client(s), which can be used in lieu of the global
initialization.

Metadata is fetched from Rune's GraphQL API
(https://graph.runelabs.io/graphql), using a
:class:`~runeq.resources.client.GraphClient`. Timeseries data is fetched
from the `V2 Stream API <https://docs.runelabs.io/stream/v2/index.html>`_,
using a :class:`~runeq.resources.client.StreamClient`.

"""
from . import client, common, org, patient, stream, stream_metadata, user

__all__ = ["client", "common", "org", "patient", "stream_metadata", "stream", "user"]



"""
Tests for common base classes.

"""
from typing import List, Type
from unittest import TestCase

from runeq.resources.common import ItemBase, ItemSet
from runeq.resources.org import OrgSet


class StreamItem(ItemBase):
    """
    Test ItemBase class that represents a Stream

    """

    def __init__(self, id: str, **attributes):
        """
        Initializes StreamItem

        """
        super().__init__(id=id, **attributes)


class PatientItem(ItemBase):
    """
    Test ItemBase class that represents a Patient

    """

    def __init__(self, id: str, name: str = "", **attributes):
        """
        Initializes PatientItem

        """
        self.name = name
        super().__init__(id=id, name=name, **attributes)


class PatientItemSet(ItemSet):
    """
    Test ItemSet class that represents a Patient

    """

    def __init__(self, items: List[PatientItem] = []):
        """
        Initializes PatientItemSet

        """
        super().__init__(items=items)

    @property
    def _item_class(self) -> Type[ItemBase]:
        """
        Instance type of items in this set.

        """
        return PatientItem


class TestItemBase(TestCase):
    """
    Unit tests for the common ItemBase class.

    """

    def test_eq(self):
        """
        Test __eq__

        """
        patient1 = PatientItem(id="patient_id_1")
        patient2 = PatientItem(id="patient_id_2")

        stream1 = StreamItem(id="stream_id_1")
        stream_with_patient1_id = StreamItem(id="patient_id_1")

        # Reflexive property
        self.assertEqual(patient1, patient1)

        # Same type different id
        self.assertNotEqual(patient1, patient2)

        # Different type and different id
        self.assertNotEqual(patient1, stream1)

        # Different type same id
        self.assertNotEqual(patient1, stream_with_patient1_id)

    def test_get(self):
        """
        Test __getattr__, __getitem__, and get

        """
        patient1 = PatientItem(id="patient_id", key1="val1")

        # __getitem__
        self.assertEqual("val1", patient1["key1"])

        # get
        self.assertEqual("val1", patient1.get("key1"))
        self.assertEqual(None, patient1.get("invalid_key"))
        self.assertEqual(0, patient1.get("invalid_key", 0))

    def test_repr(self):
        """
        Test __repr__

        """
        # Class that has a name attribute
        patient = PatientItem(id="patient_id1", name="Patient 1", key1="val1")
        self.assertEqual(
            'PatientItem(id="patient_id1", name="Patient 1")', repr(patient)
        )

        # Class that doesn't have a name attribute
        stream = StreamItem(
            id="stream-123",
        )
        self.assertEqual('StreamItem(id="stream-123")', repr(stream))

    def test_to_dict(self):
        """
        Test to_dict

        """
        patient1 = PatientItem(id="patient_id")
        self.assertEqual({"id": "patient_id", "name": ""}, patient1.to_dict())

        patient2 = PatientItem(id="patient_id", name="Patient 2", key1="val1", key2=2)
        self.assertEqual(
            {"id": "patient_id", "name": "Patient 2", "key1": "val1", "key2": 2},
            patient2.to_dict(),
        )


class TestItemSet(TestCase):
    """
    Unit tests for the common ItemSet class.

    """

    def setUp(self) -> None:
        """
        Test setup

        """
        self.test_patient_set = PatientItemSet(
            items=[
                PatientItem(id="patient1_id", name="Patient 1"),
                PatientItem(id="patient2_id", key1="val1"),
                PatientItem(id="patient3_id", key1=1, key2="val2"),
            ]
        )

    def test_iter(self):
        """
        Test __iter__

        """
        exp_patient_ids = ["patient1_id", "patient2_id", "patient3_id"]
        for patient in self.test_patient_set:
            self.assertIn(patient.id, exp_patient_ids)
            exp_patient_ids.remove(patient.id)

        self.assertEqual([], exp_patient_ids)

    def test_len(self):
        """
        Test __len__

        """
        self.assertEqual(0, len(PatientItemSet()))
        self.assertEqual(3, len(self.test_patient_set))

    def test_get(self):
        """
        Test __getitem__ and get

        """
        # __getitem__
        self.assertEqual(
            PatientItem(id="patient1_id"), self.test_patient_set["patient1_id"]
        )

        # get
        self.assertEqual(
            PatientItem(id="patient2_id", key1="val1"),
            self.test_patient_set.get("patient2_id"),
        )

    def test_add(self):
        """
        Test add

        """
        patient1 = PatientItem(id="patient1_id")
        patient2 = PatientItem(id="patient2_id")

        # Test successfully adding a new patient to a set
        patient_set = PatientItemSet([patient1])
        self.assertEqual(1, len(patient_set))

        patient_set.add(patient2)
        self.assertEqual(2, len(patient_set))
        self.assertEqual({"patient1_id", "patient2_id"}, set(patient_set.ids()))

        # Test raises TypeError when adding item with a different
        # type to the set.
        patient_set = PatientItemSet()
        new_stream = StreamItem(id="stream1_id")

        with self.assertRaisesRegex(TypeError, "must be type PatientItem"):
            patient_set.add(new_stream)

    def test_update(self):
        """
        Test update

        """
        patient1 = PatientItem(id="patient1_id")
        patient2 = PatientItem(id="patient2_id")

        # Test update with a list
        patient_set = PatientItemSet()
        patient_set.update([patient1, patient1, patient2])
        self.assertEqual(2, len(patient_set))
        self.assertEqual(patient1, patient_set.get("patient1_id"))
        self.assertEqual(patient2, patient_set.get("patient2_id"))

        # Test update with a ItemSet of the same type
        patient_set1 = PatientItemSet([patient1])
        patient_set2 = PatientItemSet([patient2])
        patient_set1.update(patient_set2)

        self.assertEqual(2, len(patient_set1))
        self.assertEqual(patient1, patient_set.get("patient1_id"))
        self.assertEqual(patient2, patient_set.get("patient2_id"))

        # other patient set is unaffected
        self.assertEqual(1, len(patient_set2))

        # Test raises TypeError when adding ItemSets with different types
        with self.assertRaisesRegex(TypeError, "cannot update with PatientItem"):
            OrgSet().update(patient_set)

    def test_remove(self):
        """
        Test remove

        """
        # Remove a patient by object or id
        patient1 = PatientItem(id="patient1_id")
        patient_set = PatientItemSet(
            items=[
                patient1,
                PatientItem(id="patient2_id", key1="val1"),
            ]
        )
        self.assertEqual(2, len(patient_set))

        patient_set.remove(patient1)
        self.assertEqual(1, len(patient_set))

        patient_set.remove("patient2_id")
        self.assertEqual(0, len(patient_set))

        # Remove multiple patients by object or id
        patient1 = PatientItem(id="patient1_id")
        patient_set = PatientItemSet(
            items=[
                patient1,
                PatientItem(id="patient2_id", key1="val1"),
            ]
        )
        self.assertEqual(2, len(patient_set))

        patient_set.remove(patient1, "patient2_id")
        self.assertEqual(0, len(patient_set))

    def test_ids(self):
        """
        Test ids

        """
        self.assertEqual([], list(PatientItemSet().ids()))

        self.assertEqual(
            ["patient1_id", "patient2_id", "patient3_id"],
            list(self.test_patient_set.ids()),
        )

    def test_repr(self):
        """
        Test __repr__

        """
        self.assertEqual(
            """PatientItemSet {
}""",
            repr(PatientItemSet()),
        )

        patient_set = PatientItemSet(
            items=[
                PatientItem(id="patient1_id", name=""),
                PatientItem(id="patient2_id", name="Patient 2"),
                PatientItem(id="patient3_id", name="Patient 3", key1="value1"),
                PatientItem(id="patient4_id"),
                PatientItem(id="patient5_id"),
            ]
        )

        self.assertEqual(
            (
                "PatientItemSet {\n"
                '	PatientItem(id="patient1_id", name="")\n'
                '	PatientItem(id="patient2_id", name="Patient 2")\n'
                '	PatientItem(id="patient3_id", name="Patient 3")\n'
                "	... (and 2 others)\n"
                "}"
            ),
            str(patient_set),
        )

    def test_to_list(self):
        """
        Test to_list

        """
        self.assertEqual([], PatientItemSet().to_list())
        self.assertEqual(
            [
                {"id": "patient1_id", "name": "Patient 1"},
                {"id": "patient2_id", "key1": "val1", "name": ""},
                {"id": "patient3_id", "key1": 1, "key2": "val2", "name": ""},
            ],
            self.test_patient_set.to_list(),
        )



from setuptools import find_packages, setup

package_name = "runeq"

with open("README.md") as r:
    long_description = r.read()

with open("requirements/common.txt", "r") as r:
    install_requires = r.read().splitlines()

setup(
    name=package_name,
    version="0.14.0",
    author="Rune Labs",
    maintainer_email="support@runelabs.io",
    description="Query data from Rune Labs APIs",
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords="rune labs api query data",
    url="https://github.com/rune-labs/runeq-python",
    packages=find_packages(),
    install_requires=install_requires,
    test_suite="tests",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
    ],
    entry_points={
        "console_scripts": [
            "runeq=cli:cli",
        ],
    },
)



"""
Query data from the Rune Labs Stream API (V1).

"""

import csv
from logging import getLogger
from typing import Iterator, Union
from urllib.parse import urlencode, urljoin

import requests

from runeq import config
from runeq.errors import APIError

try:
    import numpy as np

    USE_NUMPY = True
except ImportError:
    USE_NUMPY = False

log = getLogger(__name__)


def _str2float(s: str) -> Union[float, str]:
    """
    Convert string to a numeric type, if possible.

    """
    try:
        # if numpy is installed, use np.float64 instead of Python built-in
        if USE_NUMPY:
            return np.float64(s)
        else:
            return float(s)
    except ValueError:
        return s


def _check_response(r: requests.Response) -> None:
    """
    Raise an exception if the request was not successful.

    """
    if r.ok:
        return

    # When possible, the API returns details about what went wrong in
    # the json body of the response. Incorporate that detail into the error
    # raised.
    try:
        data = r.json()
    except Exception:
        r.raise_for_status()
        return

    if "error" in data:
        raise APIError(r.status_code, data["error"])


class StreamV1Base:
    """
    Base class for requesting data from the Stream V1 API.

    """

    # Name of the resource (e.g. lfp, accel)
    _resource: str

    # Expression to use to query the availability timeseries for a resource.
    # Not supported for all resources.
    _availability = None

    def __init__(self, cfg: config.BaseConfig, **defaults):
        """
        Initialize with a Config and default query parameters.

        Args:
            config: Config, with settings used for all API requests
            **defaults: Default query parameters to use for all API requests.
                If a method accepts query parameters, they will be used to
                override these defaults for the related requests.

        :meta public:
        """
        self.config = cfg
        self.defaults = defaults

    def _update_params(self, params):
        """
        Update query params in place, using self.defaults.

        """
        for k in filter(lambda x: x not in params, self.defaults):
            params[k] = self.defaults[k]
        return params

    def get_json_response(self, **params) -> requests.Response:
        """
        Make a GET request to the resource's JSON endpoint.

        Args:
            **params: Query parameters for the request. These override
                self.defaults, on a key-by-key basis, for this request.

        Returns:
            requests.Response

        """
        self._update_params(params)

        url = urljoin(self.config.stream_url, "/v1/{}.json".format(self._resource))
        log.debug(f"GET {url}?{urlencode(params)}")
        return requests.get(
            url,
            headers=self.config.auth_headers,
            params=params,
        )

    def iter_json_data(self, **params) -> Iterator[dict]:
        """
        Iterate over results from the resource’s JSON endpoint.

        Follows pagination to get a complete set of results, starting
        with the page specified in the `page` kwarg (or the first
        page, by default).

        Args:
            **params: Query parameters for the request. These override
                self.defaults, on a key-by-key basis.

        Yields:
             dict with the "result" from the JSON body of each response.

        Raises:
            APIError: when a request fails
        """
        has_next = True
        while has_next:
            r = self.get_json_response(**params)
            _check_response(r)
            data = r.json()

            next_page = data.get("next_page")
            next_page_token = r.headers.get("X-Rune-Next-Page-Token")
            has_next = next_page or next_page_token

            if next_page_token is not None:
                params["next_page_token"] = next_page_token
                params.pop("page", None)
            else:
                params["page"] = next_page
                params.pop("next_page_token", None)

            yield data["result"]

    @property
    def expr_availability(self):
        """
        Availability expression for this resource.

        Raises:
            NotImplementedError: if the resource doesn't support an
                availability query expression.

        """
        if not self._availability:
            raise NotImplementedError(
                f"{self.__class__.__name__} does not " f"support an availability query"
            )

        return self._availability

    def iter_json_availability(self, **params) -> Iterator[dict]:
        """
        Convenience method to query the JSON endpoint, using the availability
        expression. This may not be supported for all endpoints.

        Args:
            **params: Query parameters for the request. These override
                self.defaults, on a key-by-key basis.

        Yields:
             dict with the "result" from the JSON body of each response.

        Raises:
            NotImplementedError: if the resource doesn't support an
                availability query expression.
        """
        params["expression"] = self.expr_availability
        yield from self.iter_json_data(**params)


class StreamV1CSVBase(StreamV1Base):
    """
    Base class for requesting data from the Stream V1 API, for resources that
    support both JSON and CSV data formats.

    """

    def get_csv_response(self, **params) -> requests.Response:
        """
        Make a GET request to the resource's CSV endpoint.

        Args:
            **params: Query parameters for the request. These override
                self.defaults, on a key-by-key basis.

        Returns:
            requests.Response
        """
        self._update_params(params)

        url = urljoin(self.config.stream_url, "/v1/{}.csv".format(self._resource))

        log.debug(f"GET {url}?{urlencode(params)}")
        return requests.get(
            url,
            headers=self.config.auth_headers,
            params=params,
            stream=True,
        )

    def iter_csv_text(self, **params) -> Iterator[str]:
        """
        Iterate over CSV text results, from the resource's CSV endpoint.

        Follows pagination to get a complete set of results, starting
        with the page specified in the `page` kwarg (or the first
        page, by default).

        Args:
            **params: Query parameters for the request. These override
                self.defaults, on a key-by-key basis.

        Yields:
             text body of each response.

        Raises:
            APIError: when a request fails
        """

        # page will be maintained independently of the params value
        # to keep the page synchronized with the offset when the next
        # page token header field is present. This will allow the page
        # parameter to be added back to the params in the event that
        # the next page token header field is absent from the response.
        page = 0

        if "page" in params:
            page = params["page"]

        while True:
            r = self.get_csv_response(**params)
            _check_response(r)
            if not r.text:
                return

            yield r.text

            page += 1
            next_page_token = r.headers.get("X-Rune-Next-Page-Token")

            if next_page_token is not None:
                params["next_page_token"] = next_page_token
                params.pop("page", None)
            else:
                params["page"] = page
                params.pop("next_page_token", None)

    def iter_csv_availability(self, **params) -> Iterator[dict]:
        """
        Convenience method to query the CSV endpoint, using the availability
        expression. May not be supported for all resources.

        Args:
            **params: Query parameters for the request. These override
                self.defaults, on a key-by-key basis.

        Yields:
             text body of each response.

        Raises:
            NotImplementedError: if the resource doesn't support an
                availability query expression.
        """
        params["expression"] = self.expr_availability
        yield from self.iter_csv_text(**params)

    def points(self, **params) -> Iterator[dict]:
        """
        Iterate over points from CSV response, yielding dictionaries.

        This may involve multiple requests to the CSV endpoint, to follow
        pagination.

        Args:
            **params: query parameters for the request(s). These override
                self.defaults on a key-by-key basis.

        Yields:
             dict: Keys are the headers from the CSV response. Values are
             converted to numeric types where applicable. If numpy
             is available, np.float64 is used.

        """
        if USE_NUMPY:
            restval = np.NaN
        else:
            restval = None

        for body in self.iter_csv_text(**params):
            reader = csv.DictReader(body.splitlines(), restval=restval)
            for point in reader:
                for k in point:
                    if k is None:
                        # If a data row has more items than the header row,
                        # DictReader adds the overflow to a None key. This is
                        # unexpected behavior from the v1 API; log a warning,
                        # but return the data anyway (unconverted).
                        log.warning("Data row had too many values")
                    else:
                        if point[k] == "":
                            point[k] = restval
                        else:
                            point[k] = _str2float(point[k])

                yield point

    def __iter__(self) -> Iterator[dict]:
        """
        Iterate over points from the CSV response, using self.points().

        Yields:
            dict for each line in the CSV response. Keys are the CSV headers.

        :meta public:
        """
        yield from self.points()


####################
# V1 API Resources #
####################


class Accel(StreamV1CSVBase):
    """
    Query accelerometry data streams.

    """

    _resource = "accel"
    _availability = "availability(accel)"


class BandPower(StreamV1CSVBase):
    """
    Query band power data streams.

    """

    _resource = "band_power"
    _availability = "availability(power)"


class Event(StreamV1Base):
    """
    Query patient events.

    """

    _resource = "event"


class HeartRate(StreamV1CSVBase):
    """
    Query heart rate data streams.

    """

    _resource = "heartrate"
    _availability = "availability(bpm)"


class LFP(StreamV1CSVBase):
    """
    Query local field potential (LFP) data streams.

    """

    _resource = "lfp"
    _availability = "availability(lfp)"


class ProbabilitySymptom(StreamV1CSVBase):
    """
    Query the probability of a symptom.

    """

    _resource = "probability_symptom"
    _availability = "availability(probability)"


class Rotation(StreamV1CSVBase):
    """
    Query rotation data streams.

    """

    _resource = "rotation"
    _availability = "availability(rotation)"


class Span(StreamV1Base):
    """
    Query spans.

    """

    _resource = "span"

    def iter_json_data(self, **params) -> Iterator[dict]:
        """
        Iterate over results from the resource’s JSON endpoint.

        Follows pagination to get a complete set of results.

        Args:
            **params: Query parameters for the request. These override
                self.defaults, on a key-by-key basis.

        Yields:
             dict with the "result" from the JSON body of each response.

        Raises:
            APIError: when a request fails
        """
        # Span endpoint only supports the next page token pagination
        next_page_token = 1  # init with something truthy
        while next_page_token:
            r = self.get_json_response(**params)
            _check_response(r)
            data = r.json()

            next_page_token = r.headers.get("X-Rune-Next-Page-Token")
            if next_page_token is not None:
                params["next_page_token"] = next_page_token

            yield data["result"]


class State(StreamV1CSVBase):
    """
    Query device state.

    """

    _resource = "state"
    _availability = "availability(state)"


####################
# V1Client Factory #
####################


class V1Client:
    """
    V1Client is a factory class. It holds configuration, which is used to
    initialize accessor classes for Stream V1 endpoints.

    """

    def __init__(self, cfg: config.Config):
        """
        Initialize client with a config.

        """
        self._cfg = cfg

    @property
    def config(self) -> config.Config:
        """
        Return configuration.

        """
        return self._cfg

    def Accel(self, **defaults) -> Accel:
        """
        Initialize an Accel accessor.

        Args:
            **defaults: Default query parameters

        """
        return Accel(self.config, **defaults)

    def BandPower(self, **defaults) -> BandPower:
        """
        Initialize a BandPower accessor.

        Args:
            **defaults: Default query parameters

        """
        return BandPower(self.config, **defaults)

    def Event(self, **defaults) -> Event:
        """
        Initialize an Event accessor.

        Args:
            **defaults: Default query parameters

        """
        return Event(self.config, **defaults)

    def HeartRate(self, **defaults) -> HeartRate:
        """
        Initialize a HeartRate accessor.

        Args:
            **defaults: Default query parameters

        """
        return HeartRate(self.config, **defaults)

    def LFP(self, **defaults) -> LFP:
        """
        Initialize an LFP accessor.

        Args:
            **defaults: Default query parameters

        """
        return LFP(self.config, **defaults)

    def ProbabilitySymptom(self, **defaults) -> ProbabilitySymptom:
        """
        Initialize a ProbabilitySymptom accessor.

        Args:
            **defaults: Default query parameters

        """
        return ProbabilitySymptom(self.config, **defaults)

    def Rotation(self, **defaults) -> Rotation:
        """
        Initialize a Rotation accessor.

        Args:
            **defaults: Default query parameters

        """
        return Rotation(self.config, **defaults)

    def Span(self, **defaults) -> Span:
        """
        Initialize a Span accessor.

        Args:
            **defaults: Default query parameters

        """
        return Span(self.config, **defaults)

    def State(self, **defaults) -> State:
        """
        Initialize a State accessor.

        Args:
            **defaults: Default query parameters

        """
        return State(self.config, **defaults)



"""
Configuration for accessing Rune APIs.

"""

import os

import boto3
import yaml

AUTH_METHOD_CLIENT_KEYS = "client_keys"
"""
Auth method to use a Client Key pair for authentication.

"""

AUTH_METHOD_JWT = "jwt"
"""
Auth method to use a JWT for authentication.

"""

AUTH_METHOD_COGNITO_REFRESH = "cognito_refresh"
"""
Auth method to use a Cognito refresh token for authentication.

"""

AUTH_METHOD_ACCESS_TOKEN = "access_token"
"""
Auth method to use a user access token for authentication.

"""

DEFAULT_CONFIG_YAML = "~/.rune/config"
"""
Default path for the config file (yaml formatted)

"""


class BaseConfig:
    """Base class to hold configuration for accessing Rune APIs."""

    graph_url: str
    stream_url: str

    def refresh_auth(self) -> bool:
        """
        Refresh authentication credentials if possible, returning a bool
        indicating if the refresh occurred.

        This is specific to the authentication style: e.g. it may
        be implemented to refresh a JWT. By default, it is a no-op.

        The API clients contain logic to catch possible authentication
        errors, invoke this method, and retry the request (if credentials
        are successfully refreshed).

        """
        return False

    @property
    def auth_headers(self) -> dict:
        """
        Authentication headers for HTTP requests to Rune APIs.

        """
        raise NotImplementedError("auth_headers must be implemented by a subclass")


class Config(BaseConfig):
    """
    Holds configuration for Rune API usage (e.g. auth credentials, URLs, etc)

    """

    def __init__(self, *args, **kwargs):
        """
        Initialize configuration options.

        Args:
            *args: Accepts at most 1; a filename. If provided, values will be
                loaded from the file, using `load_yaml()`. It is invalid to
                provide both a filename **and** keyword arguments.
            **kwargs: Passed to set_values().

        Examples:
            There are three valid ways to create a config:

            >>> cfg = Config()
            # Load from default file location (~/.rune/config)

            >>> cfg = Config('./example_config.yaml')
            # Load from a specified YAML file

            >>> cfg = Config(access_token_id='foo', access_token_secret='bar')
            # Set values using keyword arguments. This can be used with any
            # valid combination of config options; the example above sets a
            # user access token.

        """
        self.graph_url = "https://graph.runelabs.io"
        self.stream_url = "https://stream.runelabs.io"
        self.auth_method = None

        self._access_token_id = None
        self._access_token_secret = None
        self._client_access_key = None
        self._client_key_id = None
        self._jwt = None
        self._cognito_client_id = None
        self._cognito_refresh_token = None

        if args and kwargs:
            raise TypeError(
                "__init__() cannot accept both positional arguments and "
                "keyword arguments"
            )
        elif args:
            if len(args) > 1:
                raise TypeError(
                    "__init__() takes at most 1 positional argument but "
                    f"{len(args)} were given"
                )

            self.load_yaml(args[0])
        elif not kwargs:
            self.load_yaml()
        else:
            self.set_values(**kwargs)

    def load_yaml(self, filename=DEFAULT_CONFIG_YAML):
        """
        Set values from a YAML file. Keys from the file are passed directly to
        set_values(), as kwargs.

        Args:
            filename: File path for a YAML-formatted config file

        """
        with open(os.path.expanduser(filename)) as f:
            params = yaml.safe_load(f)
        return self.set_values(**params)

    def set_values(
        self,
        auth_method=None,
        access_token_id=None,
        access_token_secret=None,
        client_key_id=None,
        client_access_key=None,
        jwt=None,
        cognito_client_id=None,
        cognito_refresh_token=None,
        cognito_region_name="us-west-2",
        stream_url=None,
        graph_url=None,
        **kwargs,
    ):
        """
        Set configuration values.

        Args:
            auth_method: What type of authentication credentials to use. Must
                be one of 'access_token', 'client_keys', or 'jwt'. If not set,
                the auth method is inferred based on which credentials are
                specified (as long as it's unambiguous).
            access_token_id: User access token ID
            access_token_secret: User access token secret
            client_key_id: Client key ID
            client_access_key: Client access key
            jwt: JWT
            stream_url: base URL to use for the stream API
            graph_url: base URL to use for the graph API
            **kwargs: Arbitrary values may be provided, but they are ignored.

        """
        if stream_url is not None:
            self.stream_url = stream_url

        if graph_url is not None:
            self.graph_url = graph_url

        if auth_method is None:
            num_auth_methods_set = sum(
                [
                    bool(access_token_id or access_token_secret),
                    bool(client_access_key or client_key_id),
                    bool(jwt),
                    bool(cognito_refresh_token and cognito_client_id),
                ]
            )

            if num_auth_methods_set > 1:
                raise ValueError(
                    "Cannot infer auth method: multiple credentials were "
                    "provided. Specify auth_method kwarg to disambiguate."
                )
            elif access_token_id and access_token_secret:
                auth_method = AUTH_METHOD_ACCESS_TOKEN
            elif client_key_id and client_access_key:
                auth_method = AUTH_METHOD_CLIENT_KEYS
            elif jwt:
                auth_method = AUTH_METHOD_JWT
            elif cognito_refresh_token and cognito_client_id:
                auth_method = AUTH_METHOD_COGNITO_REFRESH
            else:
                raise ValueError(
                    "Cannot infer auth method: a complete set of credentials "
                    "was not provided."
                )

        self.auth_method = auth_method

        if access_token_id is not None:
            self._access_token_id = access_token_id

        if access_token_secret is not None:
            self._access_token_secret = access_token_secret

        if client_key_id is not None:
            self._client_key_id = client_key_id

        if client_access_key is not None:
            self._client_access_key = client_access_key

        if jwt is not None:
            self._jwt = jwt

        if cognito_client_id is not None:
            self._cognito_client_id = cognito_client_id

        if cognito_refresh_token is not None:
            self._cognito_refresh_token = cognito_refresh_token

        self._cognito_client = boto3.client(
            "cognito-idp",
            region_name=cognito_region_name,
        )

        # access auth headers to ensure they're valid
        _ = self.auth_headers

    @property
    def client_auth_headers(self):
        """
        Authentication headers for HTTP requests, using client key pair.

        """
        if not self._client_access_key:
            raise ValueError("Client access key is not set")
        if not self._client_key_id:
            raise ValueError("Client key id is not set")

        return {
            "X-Rune-Client-Key-ID": self._client_key_id,
            "X-Rune-Client-Access-Key": self._client_access_key,
        }

    @property
    def jwt_auth_headers(self):
        """
        Authentication headers for HTTP requests, using a JWT.

        """
        if not self._jwt:
            if not self.refresh_auth():
                raise ValueError("JWT is not set")

        return {
            "X-Rune-User-Access-Token": self._jwt,
        }

    def refresh_auth(self) -> bool:
        """
        Use the refresh token to get a new JWT.

        Returns:
             Bool indicating if JWT was refreshed.

        Raises:
            Exception if the refresh token request fails

        """
        if not self._cognito_refresh_token:
            return False

        response = self._cognito_client.initiate_auth(
            AuthFlow="REFRESH_TOKEN_AUTH",
            AuthParameters={
                "REFRESH_TOKEN": self._cognito_refresh_token,
            },
            ClientId=self._cognito_client_id,
        )
        self._jwt = response["AuthenticationResult"]["AccessToken"]
        return True

    @property
    def access_token_auth_headers(self):
        """
        Authentication headers for HTTP requests, using a User access token.

        """
        if not self._access_token_id:
            raise ValueError("Access token id is not set")

        if not self._access_token_secret:
            raise ValueError("Access token secret is not set")

        return {
            "X-Rune-User-Access-Token-Id": self._access_token_id,
            "X-Rune-User-Access-Token-Secret": self._access_token_secret,
        }

    @property
    def auth_headers(self):
        """
        Authentication headers for HTTP requests to Rune APIs.

        """
        if self.auth_method == AUTH_METHOD_ACCESS_TOKEN:
            return self.access_token_auth_headers
        elif self.auth_method == AUTH_METHOD_CLIENT_KEYS:
            return self.client_auth_headers
        elif self.auth_method == AUTH_METHOD_JWT:
            return self.jwt_auth_headers
        elif self.auth_method == AUTH_METHOD_COGNITO_REFRESH:
            return self.jwt_auth_headers
        else:
            raise ValueError(
                f'Invalid auth_method "{self.auth_method}": expected one of '
                f"({AUTH_METHOD_ACCESS_TOKEN}, "
                f"{AUTH_METHOD_CLIENT_KEYS}, "
                f"{AUTH_METHOD_JWT}, "
                f"{AUTH_METHOD_COGNITO_REFRESH})"
            )
